// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of a logic block terminal. A Terminal object 
contains information about the logic block that it belongs to 
and the net that it drives or is driven by. Further, a Terminal
object is also associated with a type, in case the user needs the
option of creating terminals of different types. Note that the
type is not an indicator of whether the terminal is an input or
output terminal of the logic block that it is associated with. 
The input / output information is separately stored in 'isSource'
and 'isSink' fields. Finally, a Terminal object also stores a
pointer to the source/sink routing node that it is currently
mapped to. 
****************************************************************/    

#ifndef TERMINAL_H
#define TERMINAL_H

#include "logicBlock.h"
#include "Net.h"
#include "vprArch.h"

class Terminal{
 public:
  //char name [MAX_STR_LEN];  //name of terminal
  string name;  //name of terminal
  logicBlock* lBlock;  //logic block to which this terminal belongs
  Net* net;  //the net that this terminal drives or is driven by
  routingNode* rNode;  //the routing node that this terminal is currently mapped to
  enum e_terminalType terminalType;  //type of terminal. User can define types in error.h
  bool isSink;  //whether this is a sink terminal of net
  bool isSource;  //whether this is the source terminal of net
  int magicNum;

  //Timing-driven stuff
  //If this is an input terminal, it will have a list of output terminals that
  //it fans out to
  list <Terminal*> *fanOut;

  //If this is an output terminal, it will have a list of input terminals that 
  //it depends on
  list <Terminal*> *dependency;

  //If this terminal is an input it will have an arrival time
  double arrivalTime;

  //If this terminal is an output it will have a departure time
  double departureTime;

  //Both 
  double requiredTime;

  //Pipeline Routing Stuff
  //If this terminal takes part in a pipelined net, we will have non-null
  //values for one or both of these pointers.  Otherwise, just use the normal net value.

  //This is the pipelining net that drives this delay node
  PipeliningNet* sourcePipeliningNet;
  //This is the pipelining net that is driven by this delay node
  PipeliningNet* destPipeliningNet;
  bool isRouted;

  double externalDelay;
  double upstreamCost;

  // These get the (x,y) coordinates of the mapped logic block or delay block
  int getLogicBlockCol(){
    //If this terminal is mapped to a logic block, return the column coordinate.
    //If not, see if it is mapped to a delay block   
    if(lBlock){
      return(lBlock->col);
    }
    else if(rb){
      return (rb->getHostCLB()->get_col());
    }
    else{
      ERROR("Trying to access coordinate of un-mapped logic block");
    }
  }

  int getLogicBlockRow(){
    //If this terminal is mapped to a logic block, return the row coordinate.
    //If not, see if it is mapped to a delay block
    if(lBlock){
      return(lBlock->row);
    }
    else if(rb){
      return (rb->getHostCLB()->get_row());
    }
    else{
      ERROR("Trying to access coordinate of un-mapped logic block");
    }
  }

  // If this is a register, what architecture feature is it mapped to?
  registerBank *rb;

  //If this terminal is a pipelining register, has it been temporarily
  // removed from the netlist to try out a move?
  bool retimeNerfed;

  void setRegisterBank(registerBank *r){
    rb = r;
  }
  registerBank* getRegisterBank(){
    return rb;
  }

/*   void set_delayBlock(delayBlock *block, int b){ */
/*     db = block;  */
/*     bank = b; */
/*   }; */

/*   delayBlock* get_delayBlock(){ */
/*     return db; */
/*   } */

/*   int get_bank(){ */
/*     return bank; */
/*   } */

  Terminal(){
    lBlock=0;
    net=0;
    rNode=0;
    terminalType=T_NO_TYPE;
    isSink=isSource=false;
    magicNum=16111;
    
    fanOut = 0;
    dependency = 0;

    arrivalTime = -100;
    departureTime = -100;
    requiredTime = -100;
    
    sourcePipeliningNet = 0;
    destPipeliningNet = 0;
    isRouted = false;

    externalDelay = 0.0;
    upstreamCost = 0.0;
    retimeNerfed = false;

    rb = 0;
  }

  ~Terminal(){
    if(magicNum!=16111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  Terminal(string name, logicBlock* lb, Net* tnet, enum e_terminalType term_type, bool is_source){
    isSink=isSource=false;
    rNode=0;

    magicNum=16111;
    
    fanOut = 0;
    dependency = 0;
    
    arrivalTime = -100;
    departureTime = -100;
    requiredTime = -100;
    
    sourcePipeliningNet = 0;
    destPipeliningNet = 0;
    isRouted = false;
    
    externalDelay = 0.0;
    upstreamCost = 0.0;
    
    set_name(name);
    set_lBlock(lb);
    set_net(tnet);
    set_terminalType(term_type);
    if(is_source)
      set_isSource(true);
    else
      set_isSink(true);
    retimeNerfed = false;

    rb = 0;
  }

  Terminal(Net* n, routingNode* r,  PipeliningNet* sN,  PipeliningNet* dN, string nameStr){
    lBlock=0;
    net = n;
    rNode = r;
    terminalType=T_NO_TYPE;
    isSink=isSource=false;
    magicNum=16111;

    fanOut = 0;
    dependency = 0;

    arrivalTime = -100;
    departureTime = -100;
    requiredTime = -100;

    sourcePipeliningNet = sN;
    destPipeliningNet = dN; 
    //strcpy(name,nameStr);
    name = nameStr;

    externalDelay = 0.0;
    upstreamCost = 0.0;
    retimeNerfed = false;

    rb = 0;
  }
  
  Terminal(Net* n, PipeliningNet* sN,  PipeliningNet* dN, string nameStr){
    lBlock=0;
    net = n;
    terminalType=T_NO_TYPE;
    isSink=isSource=false;
    magicNum=16111;

    fanOut = 0;
    dependency = 0;

    arrivalTime = -100;
    departureTime = -100;
    requiredTime = -100;

    sourcePipeliningNet = sN;
    destPipeliningNet = dN; 
    //strcpy(name,nameStr);
    name = nameStr;

    externalDelay = 0.0;
    upstreamCost = 0.0;
    retimeNerfed = false;

    rb = 0;
  }


  void set_name(string n);  //give this terminal a name
  void set_lBlock(logicBlock* l);  //set logic block associated with this terminal
  void set_net(Net* n);  //net that this terminal drives or is driven by
  void set_terminalType(enum e_terminalType t);  //set the type of this terminal
  void set_isSink(bool u);  //whether this is a sink terminal of net
  void set_isSource(bool u);  //whether this is the source terminal of net
  void set_rNode(routingNode* r);  //set the routing node that this terminal is mapped to

  string get_name();  //get name of this terminal
  logicBlock* get_lBlock();  //get logic block associated with this terminal
  Net* get_net();  //get net that this terminal drives or is driven by
  bool get_isSink();  //whether this is a sink terminal
  bool get_isSource();  //whether this is a source terminal
  routingNode* get_rNode();  //get the routing node that this terminal is mapped to

  void clearRNode();

  void print_Info();  //print details about this terminal

  enum e_genericBlockType getHostBlockType();  

  enum e_timingMoveBlockType getHostTimingMoveType();
  
  int getHostBankNum();  

  void updateSinkDelayAndSlack();
  void updateSinkDelayAndSlackBackwards();
  void updateSourceDelaysAndSlacks();
  void updateSourceDelaysAndSlacksForwards();
  void updateSinkWireAndTimingCost(bool timing_driven, double criticalPathDelay);
  void updateSinkWireAndTimingCostBackwards(bool timing_driven, double criticalPathDelay);
  void updateSourceWireAndTimingCosts(bool timing_driven, double criticalPathDelay);
  void updateSourceWireAndTimingCostsForwards(bool timing_driven, double criticalPathDelay);
  double sinkDeltaWireCost();
  double sinkDeltaWireCostBackwards();
  double sourceDeltaWireCost();
  double sourceDeltaWireCostForwards();
  double sinkDeltaTimingCost();
  double sinkDeltaTimingCostPrime();
  double sinkDeltaTimingCostBackwards();
  double sourceDeltaTimingCost();
  double sourceDeltaTimingCostPrime();
  double sourceDeltaTimingCostForwards();
  bool hasFanOut();

  int numRegsFreedForward(int numDelays);
  int numRegsAddedBackward(int numDelays);
  int numRegsFreedBackward(int numDelays);
  bool signalAvailable(int numDelays);
  int numRegsAddedForward(int numDelays);

  PipeliningNet* detachAndReattachHereForward(Netlist *netlist, int numMove);
  PipeliningNet* detachAndReattachHereBackward(Netlist *netlist, int numMove);
  void detachAndReattachHereBackward(Terminal *sink, Netlist *netlist, int latency, bool nerfRegisters);
  bool addNRegistersBackward(Netlist *netlist, int numMove, double criticalPathDelay);
  bool addNRegistersForward(Netlist *netlist, int numMove, double criticalPathDelay);
  void reattachNetForwards(PipeliningNet *oldPNet, Netlist *netlist, int numMove);
  void reattachNetBackwards(PipeliningNet *oldPNet, Netlist *netlist, int numMove);
  void detachAndReattachHereBackwardsAborted(Netlist *netlist);
  void detachAndReattachHereBackwardAborted(Terminal *sink, Netlist *netlist);
  void detachAndReattachHereForwardsAborted(Netlist *netlist);
  void detachAndReattachHereForwardAborted(Terminal *source, Netlist *netlist);
  
  //void addBackRegisterAndNetForwards(Netlist *netlist, int numMove);
  
  void loopsBackTo(logicBlock *lb, int *latency, bool *fanOut, bool *loopsBack);

  void clearNerfedBlocksBackward();
  void clearNerfedBlocksForward();
  void checkBackwardsCapacity();
  void checkForwardsCapacity();

  void checkSourceAndDestNet();

  enum e_genericBlockType returnDriverType();
  void placeWithParent();
  bool parentPlaced();
  void printDistanceToAllSinks();
  int minDistanceToAllSinks();
};

#endif


