// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA* FPGA;
extern bool placementRetiming;
extern bool splitDelayBlocks;
extern bool unzipNets;
extern bool lut_level_placement;
extern bool clb_level_placement;
extern bool noWindowing;
extern double tdPlacementLambda;
extern double tdPlacementUpdateInterval;
extern double tdPlacementCritExponentFinal;
extern double tdPlacementCritExponent;
extern bool maxCutCost;
extern bool aveCutCost;
extern bool noWindowing;
extern double placeInnerLoop;
extern bool registerHoming;
extern double averageManhattanDistance;
extern bool lutStickiness;
extern bool deltaDelayCost;

#ifdef PARANOID_CHECK_PIPEPLACE
extern int numRegistersEstimatedFreed;
extern int numRegistersEstimatedNeeded;
#endif

int numSplitCalls = 0;
int numSplitSwapCalls = 0;
int numSplits = 0;
int numMergeCalls = 0;
int numMergeSwapCalls = 0;
int numMerges = 0;

int numAttemptedLUTMoves;
int numAttemptedRegMoves;
int numAttemptedIOMoves;
//int averageAttemptedLUTMoveManhattanDistance;
//int averageAttemptedRegMoveManhattanDistance;
//int averageAttemptedIOMoveManhattanDistance;

// Perform the pipeline-aware placement
// We assume that the netlist already has an initial placement for the
//  logic, I/O, and delay blocks
void pipeline_placement(Netlist *netlist, bool timing_driven){
  double criticalPathDelay;
  double placementCost;
  //vector <zipWidget> zipVector;
  //vector <unZipWidget> unZipVector;
  double temperature;

  //Do the inital placement
  init_pipeline_placement(netlist);
  
  //printPlacement(netlist);

  //Now, calculate the critical path for the netlist given the initial placement
  netlist->updatePlacementTiming();
  
  criticalPathDelay = calculateCriticalPathDelay(netlist, true);
  
  //Now calculate the cost of the placement
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  printf("Initial placement Wire Cost = %f\n", netlist->wireCost);
  printf("Initial placement Timing Cost = %e\n", netlist->timingCost);
  printf("Starting critical path delay = %e\n", criticalPathDelay);

  if(maxCutCost){
    printf("Initial placement Max Cut Cost = %f\n", netlist->maxCut);
  }
  if(aveCutCost){
    printf("Initial placement Ave Cut Cost = %f\n", netlist->aveCut);
  }

  //printf("Pre-initial placement placement\n");
  //netlist->printPlacement();
  //netlist->print_netList(true);

  temperature = pipelineDetermineAnnealingTemp(netlist, timing_driven, &criticalPathDelay);
  printf("Initial Temp = %e\n", temperature);

  //Start the annealing
  pipelineAnneal(netlist, timing_driven, temperature, (float) max(FPGA->get_numRows(), FPGA->get_numCols()));

  //printf("Final placement Wire Cost = %f\n", netlist->wireCost);
  //printf("Final placement Timing Cost = %e\n", netlist->timingCost);
  //printf("Final critical path delay = %e\n", criticalPathDelay);  

  //netlist->printPlacement();
  //netlist->print_netList(true);

  if(maxCutCost){
    printf("Final placement Max Cut Cost = %f\n", netlist->maxCut);
  }
  if(aveCutCost){
    printf("Final placement Ave Cut Cost = %f\n", netlist->aveCut);
  }
}


//Perform a few iterations of low-temp annealing on the existing placement
void pipelineLowTempAnneal(Netlist *netlist, bool timing_driven){
 double criticalPathDelay;
  double placementCost;
  double temperature;

  //Calculate the critical path and cost for the netlist's existing placement
  criticalPathDelay = calculateCriticalPathDelay(netlist, true);
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  temperature = 1.0E-5;
  printf("Low temp anneal starts at = %e\n", temperature);

  //Start the annealing
  pipelineAnneal(netlist, timing_driven, temperature, 1.0);
}

// Perform pipeline-aware initialization of the placement
// 1) Place all of the I/O and LUTs.
// 2) Place all register banks and delay registers based on the assumption that all nets share
//    as many pipelining registers as possible. (Fully zipped)
void init_pipeline_placement(Netlist *netlist){
  vprLUT* lut;
  vprIOB* io_block;
  registerBank *rb;
  Net* net;
  Terminal* outTerm;
  logicBlock* lb;
  Terminal* pr;
  list <logicBlock*> registerBankList;
  list <Terminal*> pipelineRegisterList;
  int counter;
  bool done;

  // Go through netlist and sort all net sinks by non-decreasing latency
  netlist->sort_netPipelineSinks(true);

  // Do the initial placement.
  // If we are doing clb-level placement just randomly place the netlist's clbs.
  // If we are doing FF-level placement, we should hold off on the placement
  //  of logic block register banks till all of the LUTs are placed.
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    //First, place the logic blocks themselves
    //Check to see if it already has been placed
    if(lb->row != -1 || lb->col != -1){
      if(clb_level_placement){
	//We can move onto the next logic block because this one has been placed already
	// because something else in the same CLB has been placed
	continue;
      }
      else{
	if(lb->get_blockType() != LUT){
	  ERROR("Logic block should not have a placement yet!");
	}
	else{
	  continue;
	}
      }
    }

    //If we are doing FF-level placement, we should hold off on the placement
    //  of logic block register banks till all of the LUTs are placed.
    //Put this register bank in the list.
    if(!clb_level_placement && lb->get_blockType() == REGISTER_BANK){
      registerBankList.push_back(lb);
      continue;
    }

    done = false;
    while(!done){
      if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
	io_block = FPGA->iobList.at(lrand48() % FPGA->iobList.size());

	//See if this io block has any empty locations
	for(int i = 0; i < io_block->get_numIOs(); i++){
	  if(!io_block->getfrom_currBlockArray(i)){
	    // Place the io pin in the empty location
	    placeBlock(lb,(void *)(io_block), IOPIN, i);
	    done = true;
	    break;
	  }
	}
      }
      else if(lb->get_blockType() == REGISTER_BANK){
	//We will only be placing register banks here if we are doing clb-level placement
	//We have to place all of the items in the same logical CLB into the same physical CLB
	rb = FPGA->getRandomFreeRegBank();
	if(rb->emptyCLB()){
	  lb->placeEntireCLB(rb);
	  done = true;
	}
      }
      else{
	if(lb->get_blockType() == LUT){
	  lut = FPGA->getRandomLUT();
	}
 	else if(lb->get_blockType() == RAM_BLOCK){
 	  lut = FPGA->getRandomRam();
 	}
 	else if(lb->get_blockType() == MULTIPLIER){
 	  lut = FPGA->getRandomMult();
 	}
	else{
	  ERROR("Unknown logic block type!");
	}
      
	//See if the logic at this location is already occupied
	if(!clb_level_placement){
	  //We are not doing clb-level placement, but we should check to make
	  // sure that this clb's luts are empty since we are still observing
	  // the LUT packing.
	  if(lut->emptyCLBLUTS()){
	    //If we are doing individual block level placement, just place all of the LUTs into the same physical CLB
	    lb->placeAllCLBLUTs(lut);
	    done = true;
	  }
	}
	else{
	  //If we are doing clb-level placement, check that the entire CLB is empty
	  if(lut->emptyCLB()){
	    //We have to place all of the items in the same CLB into the same physical location
	    lb->placeEntireCLB(lut);
	    done = true;
	  }
	}
      }
    }
  }
  netlist->remove_lbListItr();  

  // Now we have to figure out what to do with any pipelining registers that might be on the output of input blocks or 
  //   on logic block outputs
  // We assume that all signals are fully "zipped" so we have maximum sharing of any pipelining registers.
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    // This test shouldn't be necessary since output pins shouldn't have any output terminals
    if(lb->get_blockType() != OUTPUTPAD){
      lb->create_outputListItr();
      for(outTerm = lb->getnext_outputList(); outTerm != 0; outTerm = lb->getnext_outputList()){
	// See if the net associated with this output terminal is pipelined
	net = outTerm->get_net();
	if(net->pipelined){
	  if(clb_level_placement){
	    ERROR("We should not be doing CLB-level placement on unpacked netlists!");
	  }

	  // For the moment, we'll assume that all nets are fully zipped.
	  // Create pipelined nets that are fully zipped, but don't worry about how many
	  // delays we can pick up at a given delay block.  We'll worry about packing them smartly
	  // when we actually place them.
	  net->createZippedPipelinedNets();
	  net->addPipelineRegistersToList(&pipelineRegisterList);
	  netlist->addPipelinedRegistersToPlacementList(net);
	}
      }
      lb->remove_outputListItr();
    }
  }
  netlist->remove_lbListItr();
 
  if(!clb_level_placement){
    //If we are doing FF-level placement, we have not placed any register banks
    // or pipelining registers yet.
    //We'll do this in stages to maximize the likelihood that these are packed with their parent

    //First, go through the respective lists once and place any register banks or pipelining 
    // registers that are directly driven by a LUT.
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
      
      if(lb->returnDriverType() == LUT){
	//If it's parent is a LUT, place this logic block
	lb->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }

    counter = pipelineRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = pipelineRegisterList.front();
      pipelineRegisterList.pop_front();
      
      if(pr->returnDriverType() == LUT){
	//If it's parent is a LUT, place this pipelining register
	pr->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
      }
    }
   
    //Next, place any register banks or pipelining registers that are attached to other
    // register banks or pipelining registers that have placements.
    //We will go through these lists until we no long place any new items.
    done = false;
    while(!done){
      done = true;

      counter = registerBankList.size();
      for(int i = 0; i < counter; i++){
	//Grab the first item in the list
	lb = registerBankList.front();
	registerBankList.pop_front();
	
	if((lb->returnDriverType() == GBT_NO_TYPE || lb->returnDriverType() == REGISTER_BANK) && lb->parentPlaced()){
	  //If it's parent is a register of some type and it's already been placed,
	  // place this logic block.
	  lb->placeWithParent();
	}
	else{
	  //If not, add it to the back of the list;
	  registerBankList.push_back(lb);
	}
      }
      
      counter = pipelineRegisterList.size();
      for(int i = 0; i < counter; i++){
	//Grab the first item in the list
	pr = pipelineRegisterList.front();
	pipelineRegisterList.pop_front();
	
	if((pr->returnDriverType() == GBT_NO_TYPE || pr->returnDriverType() == REGISTER_BANK) && pr->parentPlaced()){
	  //If it's parent is a register of some type and it's already been placed,
	  // place this pipelining register.
	  pr->placeWithParent();
	}
	else{
	  //If not, add it to the back of the list;
	  pipelineRegisterList.push_back(pr);
	}
      }
    }

    //At this point all register banks and pipelining registers driven upstream at any point by any sort of LUT
    // should be placed.
    //We have NOT placed any registers/pipelining register that are attached to themselves (constant generators),
    // input pads, RAMs or multipliers.
    //Place constant generators.
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
      
      if(lb->isSelfFeeding()){
	//If this is a constant generator, place this logic block someplace
	rb = FPGA->getRandomFreeRegBank();
	rb->placeLogicBlock(lb);
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }
    

    //Now, go through the lists and place any register bank or pipelining register whose parent has been placed
    while(pipelineRegisterList.size() != 0 || registerBankList.size() != 0){
      counter = registerBankList.size();
      for(int i = 0; i < counter; i++){
	//Grab the first item in the list
	lb = registerBankList.front();
	registerBankList.pop_front();
	
	if(lb->parentPlaced()){
	  //If it's parent has been placed, place this logic block.
	  lb->placeWithParent();
	}
	else{
	  //If not, add it to the back of the list;
	  registerBankList.push_back(lb);
	}
      }
      
      counter = pipelineRegisterList.size();
      for(int i = 0; i < counter; i++){
	//Grab the first item in the list
	pr = pipelineRegisterList.front();
	pipelineRegisterList.pop_front();
	
	if(pr->parentPlaced()){
	  //If it's parent has been placed, place this pipelining register.
	  pr->placeWithParent();
	}
	else{
	  //If not, add it to the back of the list;
	  pipelineRegisterList.push_back(pr);
	}
      }
    }
  }
}

// We have just created a single long string of pipelining registers for this net.
// To place them, we just need to find a random empty delay block bank, then
// put as many registers as we can into it.  We might be limited by a few things:
//   1) Of course, the capacity of the bank
//   2) Equally obvious, the number of registers needed by this net
//   3) We need have a sink that needs the signal at this latency
// We can just keep mapping successive terminals into the bank until we reach one of these
// limits.  If we hit condition 1 or 3, we should stop and find another random empty delay
// block bank.  This assumes that each bank of shift registers only has one output.
// That is, for example, a RaPiD bus connector cannot provide an output at 2 delays for one signal
// and an output at 3 delays for another.
// If we want to model that, we have to figure out something else entirely.
void initialPlacePipelineRegs(Net *net){
  registerBank *rb = NULL;
  
  // This code makes the assumption that the list of pipelining registers contains
  // only one continuous string of registers (no unzipped branches) and is in order of
  // increasing latency.  Since we just created them in createZippedPipelinedNets,
  // this is safe.
  for(list<Terminal*>::iterator regTermIter = net->pipelineRegisterList.begin(); 
      regTermIter !=  net->pipelineRegisterList.end(); regTermIter++){
    if(!rb){
      // If we don't have a register bank we are working on, grab an empty one
      rb = FPGA->getRandomFreeRegBank();
    }

    //Put the current register in the bank
    rb->placeTerminal(*regTermIter);

    //We should now move onto another register bank if:
    // 1) The bank is full
    if(rb->bankFull()){
      rb = NULL;
    }
    // Or 2) If the pipeline net sourced by this register has fanout
    if((*regTermIter)->destPipeliningNet->sinkVector.size() > 1){
      rb = NULL;
    }
  }
}

void cleanUpPipelinePlacement(Netlist *netlist){
  // Go through all of the nets and remove the pipelining information 
  // we generated during placement before we do routing
  netlist->create_netListItr();
  for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
    net->cleanUpPipelinePlacementInformation();
  }
  netlist->remove_netListItr();
}

// Try some random moves to see what affect this has on the cost
// We have to return not only the temperature we should use, but also
// update the final criticalPathDelay for the system we end up with.
double pipelineDetermineAnnealingTemp(Netlist *netlist, bool timing_driven, double *criticalPathDelay){
  void* block0;
  void* block1;
  int indexOrBank0, indexOrBank1;
  int moveType;
  enum e_timingMoveBlockType blockType;
  int numMove;
  double dCost;
  vector <double> costArray;
  double average = 0.0;
  double stdDev = 0.0;
  float rLim = (float) max(FPGA->get_numRows(), FPGA->get_numCols());
  list <PipeliningNet *> retimeUndoList;
  int moves = netlist->calculateNumBlocksAndPipeliningRegisters() * 100;

  averageManhattanDistance = 0.0;

#ifdef PIPEPLACEDEBUGHEAVY
  int numberMoves = 0;
#endif

#ifdef PARANOID_CHECK_PIPEPLACE
  double oldWireCost;
  double oldTimingCost;
  double oldMaxCut;
  double oldAveCut;
#endif
  int numMovesBeforeUpdate = (int)(tdPlacementUpdateInterval * (double) moves);
  if(numMovesBeforeUpdate <= 0){
    numMovesBeforeUpdate = 1;
  }

#ifdef PARANOID_CHECK_PIPEPLACE
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  if(timing_driven){
    netlist->overallCost = 1.0;
  }
  else{
    netlist->overallCost = netlist->wireCost;
  }

  //Move each block an average of 100x
  for(int i = 0; i < moves; i++){
	
#if defined(PIPEPLACEDEBUGHEAVY)
    cout << "Wire Cost before move = " <<  netlist->wireCost <<endl;
    cout << "Timing Cost before move = " <<  netlist->timingCost <<endl;

    if(maxCutCost){
      cout << "Max Cut before move = " <<  netlist->maxCut <<endl;
    }
    if(aveCutCost){
      cout << "Ave Cut before move = " <<  netlist->aveCut <<endl;
    }
    //netlist->print_netList();
    //netlist->printPlacement();
    numberMoves ++;

    if(numberMoves == 97){
      //cout << "Before move!" << endl;
      //netlist->print_netList(true);
      //netlist->printPlacement();
    }
#endif
    
    //Generate a random move
    moveType = generatePipelineAwareMove(netlist, &block0, &block1, &indexOrBank0, &indexOrBank1, &blockType, &numMove, (int)(rLim + 0.5));

    //Make the move and update all the proper sink delays.
    //We are also making incremental changes to the slacks - they are inexact, but we'll see what
    // we can do with them
    //We don't care if the move failed, since we don't undo anything here
    makePipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, 
			  blockType, numMove, *criticalPathDelay, &retimeUndoList);

#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

    //Figure out the cost of the move
    if(deltaDelayCost){
      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
					   timing_driven, *criticalPathDelay, &retimeUndoList);
    }
    else{
      dCost = determineDeltaCost(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				 timing_driven, *criticalPathDelay, &retimeUndoList);
    }
    netlist->overallCost += dCost;
    //netlist->printPlacement();

    //If we just zipped a net, we should go ahead and delete the registers in block1 now
    // we kept them around because of incremental cost calculation and just in case
    // we had to undo the move, but now we can get rid of them
    if(moveType == MOVEZIPPNET){
      ((registerBank*)block1)->freeBank();
    }

    else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
      clearRetimeUndoListForward(&retimeUndoList);
      clearNerfedBlocksBackward(block0, blockType);
      //netlist->printPlacement();
      //netlist->print_netList();
    }

    else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
      clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
      clearNerfedBlocksForward(block0, blockType);
      //netlist->printPlacement();
      //netlist->print_netList();
    }

#if defined(PIPEPLACEDEBUGHEAVY)
    if(numberMoves == 10){
      //cout << "After move!" << endl;
      //netlist->print_netList(true);
      //netlist->printPlacement();
    }
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
    oldWireCost = netlist->wireCost;
    oldTimingCost = netlist->timingCost;
    oldMaxCut = netlist->maxCut;
    oldAveCut = netlist->aveCut;
    
    netlist->calculatePlacementCost(timing_driven, *criticalPathDelay);
    if(fabs(oldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
      ERROR("Problem doing incremental cost update");
    }
    if(fabs(oldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
      ERROR("Problem doing incremental cost update");
    }
    if(oldMaxCut != netlist->maxCut){
      ERROR("Problem doing incremental cost update");
    }
    if(oldAveCut != netlist->aveCut){
      ERROR("Problem doing incremental cost update");
    }

    //FPGA->checkFreeRegisters();
    netlist->checkNets();

    if(numRegistersEstimatedFreed != 0){
      //ERROR("Did not estimate # of freed registers correctly");
    }
    if(numRegistersEstimatedNeeded != 0){
      //ERROR("Did not estimate # of needed registers correctly");
    }
#endif

    //To be completely accurate, we should now recalculate the critical path delay and slacks
    // for the whole system.
    //However, this is too timing consuming to do after each move, so we don't do it all of the time
    // we only do it after every N moves
    if(i % numMovesBeforeUpdate == (numMovesBeforeUpdate - 1)){
      *criticalPathDelay =  calculateCriticalPathDelay(netlist, true);
	
      //Now, the costs for all of the nets might have changed, so calculate the placement cost from scratch      
      netlist->calculatePlacementCost(timing_driven, *criticalPathDelay);
    }

    //Keep track of the cost
    //costArray.push_back(netlist->overallCost);
    average += netlist->overallCost;
    stdDev += (netlist->overallCost * netlist->overallCost);

    //netlist->printPlacement();
  }

  //Calculate the average delta cost
  average /= (double)moves;
  stdDev = (stdDev - ((double)moves * average * average)) / (double)(moves);
  stdDev = pow(stdDev, 0.5);
  
  //Initial temperature is 20 standard deviations of the expected initial delta cost
  return (stdDev * 20.0);
}


//Figure out what kind of move to make. First trying to select between 
// attempting to move a logic block, (should put an I/O pin here
// but RaPiD does not allow pins to move so it's absent for the momemt), 
// grouped delay block or free-moving register.  Once we make the decision
// what kind of block we want to move, we will figure out what kind of move
// to make.
int generatePipelineAwareMove(Netlist *netlist, void** block0, void** block1, int *index0, int *index1, 
			      enum e_timingMoveBlockType *blockType, int *numMove, int rLim){
  int moveType;

  moveType = selectPipelineAwareMove(netlist);

  if(moveType == MOVELOGIC){
    generatePipelineAwareMoveLogic(&moveType, netlist, block0, block1, index0, index1, blockType, numMove, rLim);
  }
  else if(moveType == MOVEPIPEREG){
    generatePipelineAwareMovePipeReg(&moveType, netlist, block0, block1, numMove, rLim);
  }
  else{
    ERROR("Unknown Move Type!");
  }

  return (moveType);
}

int selectPipelineAwareMove(Netlist *netlist){
  long choice;

  //Figure out if we should do something to a logic block, a delay block, or a pipeline register
  // We are choosing a random number between 0 and the number of (ALU, MULT, RAM, REG) + number of delay blocks 
  //      + number of pipelining registers used in pipelined nets
  // numLogicBlocks does not include the I/O pins - we can't move them in Rapid anyways, so it's pointless
  // numDelayBlocks is the number of fixed placement delay blocks we have added into the system - it is updated 
  // whenever we decide to create one
  //   Just a note, we currently don't create any fixed delay blocks.  They are all floating pipelining registers.
  // numPipeliningRegisters is not calculated from scratch based on the current activity of the pipelined nets in the system.
  //  This means that we should keep the placementPipelineRegisterVector up-to-date when we zip or unzip registers.
  // *********************************************************
  choice = lrand48() % (netlist->get_numlogicBlocks() + netlist->getNumPipeliningRegisters());

  if(choice < netlist->get_numlogicBlocks()){
    return (MOVELOGIC);
  }
  else{
    return (MOVEPIPEREG);
  }
}

void generatePipelineAwareMoveLogic(int *moveType, Netlist *netlist, void** block0, void** block1, int *index0, int *index1, 
				    enum e_timingMoveBlockType *blockType, int *numMove, int rLim){
  logicBlock* lb;
  int row0, col0, select, choices;
  
  //Find a logic block in the netlist
  lb = netlist->getRandomLogicBlock();
  
  if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
    //If we selected an input pad, we can only do a swap.
    //No retiming moves are possible
    row0 = lb->row;
    col0 = lb->col;
    *index0 = lb->ptc_index;
    *block0 = (void*) FPGA->getIOBAt(row0, col0);

#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected block %s at %d, %d to start\n", lb->name.c_str(), ((vprIOB*)*block0)->get_col(), ((vprIOB*)*block0)->get_row());
    
#endif    
    //We are going to swap this iob with some other iob
    *block1 = *block0;
    *index1 = *index0;
    
    //Now find some other logic block in the array that matches the type of the one we picked
    while(*block0 == *block1 && *index0 == *index1){
      //*block1 = (void *) FPGA->getRandomIOB();
      *block1 = (void *) FPGA->getOtherIOB(row0, col0, rLim);
      *index1 = ((vprIOB*)*block1)->getRandomIndex();
    }

    //numAttemptedIOMoves++;
    //averageAttemptedIOMoveManhattanDistance += (abs(((vprIOB*)*block1)->get_row() - row0) + abs(((vprIOB*)*block1)->get_col() - col0));

    *blockType = IOPIN;

#if defined(PIPEPLACEDEBUGHEAVY)
    if(((vprIOB*)(*block1))->getfrom_currBlockArray(*index1)){
      printf("Selected block %s at %d, %d index %d to swap\n", ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1)->name.c_str(), 
	     ((vprIOB*)*block1)->get_col(), ((vprIOB*)*block1)->get_row(), *index1);
    }
    else{
      printf("Selected empty IOB at %d, %d index %d to swap\n", ((vprIOB*)*block1)->get_col(), 
	     ((vprIOB*)*block1)->get_row(), *index1);
    }
#endif
  }
  else if(lb->get_blockType() == REGISTER_BANK){
    //If we selected a register bank, we can do a swap or a retiming move
    row0 = lb->row;
    col0 = lb->col;
    *block0 = (void*) lb->rb;
    *blockType = REGISTERBANK;
    
#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected register bank %s at delay block %d, %d bank %d to start\n", lb->name.c_str(), ((registerBank*)*block0)->getHostCLB()->get_col(), ((registerBank*)*block0)->getHostCLB()->get_row(), ((registerBank*)*block0)->getHostBankNum());
    
    if(FPGA->getNumberofBlockType(REGISTER_BANK) == 1){
      // We might get into the situation in which we are unable to generate a move for this
      // netlist - there might be only one legal location for all of the blocks!
      // We should check for this before we get into this part of the code.
      // Can we also use this to help not make windowing make this run forever?
      
      ERROR("Cannot handle move when there is only one register bank!");
    }
#endif

    // If we are able to perform retiming moves, see if we could pull at least one set of registers
    //  from the output to the input or from the input to the output.
    if(placementRetiming){
      //Performing a normal block move will be "0"
      choices = 1;
      if(lb->numRegCanPushBackward > 0){
	//Pushing registers back will be "1"
	choices ++;
      }
      if(lb->numRegCanPushForward > 0){
	//Pushing register forward will be "2"
	choices ++;
      }
      
      select = lrand48() % choices;
      
      if(select != 0){
	if(select == 1 && lb->numRegCanPushBackward > 0){
	  //See if we can actually make any backwards retiming moves given
	  // the current state of the architecture
	  if(generatePipelineAwareMoveRegRetimeBackward(moveType, netlist, block0, *blockType, numMove)){
	    return;
	  }
	}
	else{
	  //See if we can actually more any forwards retiming moves given
	  // the current state of the architecture
	  if(generatePipelineAwareMoveRegRetimeForward(moveType, netlist, block0, *blockType, numMove)){	
	    return;
	  }
	}
      }
    }
    
    //If for whatever reason we don't want to try and retime registers through this monolithic register bank,
    // let's swap it with a different one.
    *block1 = *block0;
    
    //Now find some other register bank to swap with
    while(((registerBank*)*block0)->getHostCLB()->get_row() == ((registerBank*)*block1)->getHostCLB()->get_row() &&
	  ((registerBank*)*block0)->getHostCLB()->get_col() == ((registerBank*)*block1)->getHostCLB()->get_col()){
      //If register homing is on, then 10% of the time try to move to the same CLB as your parent LUT/register bank/pipelining register
      if(registerHoming){
	if(lrand48() % 10 == 0){
	  //Return a register bank in the same CLB as it's parent.
	  *block1 = ((registerBank*)*block0)->getRegBankInParentHostCLB();

	  //However, it might be driven by an input pad
	  if(!(*block1)){
	    *block1 = FPGA->getOtherRegBank(row0, col0, rLim);
	  }
	}
	else{
	  *block1 = FPGA->getOtherRegBank(row0, col0, rLim);
	}
      }
      else{
      //*block1 = FPGA->getRandomRegBank();
      *block1 = (void*) FPGA->getOtherRegBank(row0, col0, rLim);
      }
    }

    //numAttemptedRegMoves++;
    //averageAttemptedRegMoveManhattanDistance += (abs(((registerBank*)*block1)->getHostCLB()->get_row() - row0) + abs(((registerBank*)*block1)->getHostCLB()->get_col() - col0));

#if defined(PIPEPLACEDEBUGHEAVY)
    if(!((registerBank*)(*block1))->isBankEmpty()){
      //See if this register bank is occupied by a logic block or a pipelinine register
      if(((registerBank*)*block1)->get_currBlock()){
	printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap\n", 
	       ((registerBank*)*block1)->get_currBlock()->name.c_str(), 
	       ((registerBank*)*block1)->getHostCLB()->get_col(), 
	       ((registerBank*)*block1)->getHostCLB()->get_row(),
	       ((registerBank*)*block1)->getHostBankNum());	
      }
      else{
	printf("Selected register bank with mapped pipelining registers at delay block %d, %d bank to swap\n", 
	       ((registerBank*)*block1)->getHostCLB()->get_col(), 
	       ((registerBank*)*block1)->getHostCLB()->get_row(),
	       ((registerBank*)*block1)->getHostBankNum());	
      }
    }
    else{
      printf("Selected empty register bank at delay block %d, %d bank %d to swap\n", 
	     ((registerBank*)*block1)->getHostCLB()->get_col(), ((registerBank*)*block1)->getHostCLB()->get_row(),
	     ((registerBank*)*block1)->getHostBankNum());
    }
#endif
  }
  else if(lb->get_blockType() == LUT || lb->get_blockType() == RAM_BLOCK || lb->get_blockType() == MULTIPLIER){
    //If we selected any kind of logic block we can do a swap or a retiming move
    row0 = lb->row;
    col0 = lb->col;

    *block0 = (void*) FPGA->getCLBAt(row0, col0)->getLUT(lb->ptc_index);
    *blockType = LB;

#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected block %s at %d, %d to start\n", lb->name.c_str(), ((vprCLB*)*block0)->get_col(), ((vprCLB*)*block0)->get_row());
    
    if(FPGA->getNumberofBlockType(((vprLUT*)*block0)->get_blockType()) == 1){
      // We might get into the situation in which we are unable to generate a move for this
      // netlist - there might be only one legal location for all of the blocks!
      // We should check for this before we get into this part of the code.
      // Can we also use this to help not make windowing make this run forever?
      
      ERROR("Cannot handle move of only one block!");
    }
#endif

    // If we are able to perform retiming moves, see if we could pull at least one set of registers
    //  from the outputs to the inputs or from the inputs to the outputs.
    if(placementRetiming){
      //Performing a normal block move will be "0"
      choices = 1;
      if(lb->numRegCanPushBackward > 0){
	//Pushing registers back will be "1"
	choices ++;
      }
      if(lb->numRegCanPushForward > 0){
	//Pushing register forward will be "2"
	choices ++;
      }
      
      select = lrand48() % choices;

      if(select != 0){
	if(select == 1 && lb->numRegCanPushBackward > 0){
	  //See if we can actually make any backwards retiming moves given
	  // the current state of the architecture
	  if(generatePipelineAwareMoveRegRetimeBackward(moveType, netlist, block0, *blockType, numMove)){
	    return;
	  }
	}
	else{
	  //See if we can actually more any forwards retiming moves given
	  // the current state of the architecture
	  if(generatePipelineAwareMoveRegRetimeForward(moveType, netlist, block0, *blockType, numMove)){	
	    return;
	  }
	}
      }
    }
    
    //If for whatever reason we don't want to try and retime registers through this block,
    // let's swap it with a different one.
    
    //We are going to swap this block with some other block of the same type.
    //However, don't try to swap two blocks in the same CLB
    *block1 = *block0;

    //We should not select two blocks in the same
    // CLB either
    while(((vprLUT*)*block0)->row == ((vprLUT*)*block1)->row &&
	  ((vprLUT*)*block0)->col == ((vprLUT*)*block1)->col){
      switch(((vprLUT*)*block0)->get_blockType()){
      case LUT:
	//*block1 = (void*) FPGA->getRandomLUT();
	*block1 = (void*) FPGA->getOtherLUT(row0, col0, rLim);
	break;
      case MULTIPLIER:
	ERROR("Multipliers are not supported yet!");
	//*block1 = (void*) FPGA->getRandomMult();
	//*block1 = (void*) FPGA->getOtherMult(row0, col0, rLim);
	break;
      case RAM_BLOCK:
	ERROR("RAM blocks are not supported yet!");
	//*block1 = (void*) FPGA->getRandomRam();
	//*block1 = (void*) FPGA->getOtherRam(row0, col0, rLim);
	break;
      }
    }

    //numAttemptedLUTMoves++;
    //averageAttemptedLUTMoveManhattanDistance += (abs(((vprLUT*)*block1)->get_row() - row0) + abs(((vprLUT*)*block1)->get_col() - col0));

#if defined(PIPEPLACEDEBUGHEAVY)
    if(((vprLUT*)(*block1))->get_currBlock()){
      printf("Selected block %s at %d, %d to swap\n", ((vprLUT*)(*block1))->get_currBlock()->name.c_str(), 
	     ((vprLUT*)*block1)->get_col(), ((vprLUT*)*block1)->get_row());
    }
    else{
      printf("Selected empty block at %d, %d to swap\n", ((vprLUT*)*block1)->get_col(), ((vprLUT*)*block1)->get_row());
    }
#endif
  }
  else{
    ERROR("Unknown block type!");
  }
}


void generatePipelineAwareMovePipeReg(int *moveType, Netlist *netlist, void** block0, void** block1, int *numMove, int rLim){
  long choice;
  Terminal* term;
  
  //Find a pipelining register in the netlist
  term = netlist->getRandomPipeliningRegister();
  
  //Notice that when we select a given register above, we lose this information here.
  // We only point to a given bank in a given block.  This is because moving a single register
  // around has lots of restrictions.  For example, if there are multiple registers in a bank,
  // we cannot move any random register, only the first or last ones as a group.
  *block0 = (void*) term->getRegisterBank();
  
#if defined(PIPEPLACEDEBUGHEAVY)
  printf("\n\nSelected pipelining register %s at %d, %d bank %d to start\n", term->name.c_str(), 
	 ((registerBank*)*block0)->getHostCLB()->get_col(),
	 ((registerBank*)*block0)->getHostCLB()->get_row(),
	 ((registerBank*)*block0)->getHostBankNum());

  if(FPGA->getNumberofBlockType(REGISTER_BANK) == 1){
    // We might get into the situation in which we are unable to generate a move for this
    // netlist - there might be only one legal location for all of the blocks!
    // We should check for this before we get into this part of the code.
    ERROR("Cannot handle move of only one register bank!");
  }
#endif	

  *block1 = *block0;

  //Now find some other register bank in the array
  while(*block0 == *block1){
    //If register homing is on, then 10% of the time try to move to the same CLB as your parent LUT/register bank/pipelining register
    if(registerHoming){
      if(lrand48() % 10 == 0){
	//Return a register bank in the same CLB as it's parent.
	*block1 = ((registerBank*)*block0)->getRegBankInParentHostCLB();

	//However, it might be driven by an input pad
	if(!(*block1)){
	  *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->get_row(),
						  ((registerBank*)*block0)->getHostCLB()->get_col(), rLim);
	}
      }
      else{
	*block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->get_row(),
						((registerBank*)*block0)->getHostCLB()->get_col(), rLim);
      }
    }
    else{
      //*block1 = FPGA->getRandomRegBank();
      *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->get_row(),
					((registerBank*)*block0)->getHostCLB()->get_col(), rLim);
    }

#if defined(PIPEPLACEDEBUGHEAVY)
    printf("Selected delay block at %d, %d bank %d to swap\n", 
	   ((registerBank*)*block1)->getHostCLB()->get_col(),
	   ((registerBank*)*block1)->getHostCLB()->get_row(),
	   ((registerBank*)*block1)->getHostBankNum());
#endif
  }

  //Now that we have two particular register banks in mind, we have to figure out what to do with them.  
  if(!splitDelayBlocks){
    //If we are not splitting delay blocks, we have to swap the two
    generatePipelineAwareMovePipeRegSwap(moveType, netlist, (registerBank**) block0, 
					 (registerBank**) block1, numMove);
  }
  else if(!unzipNets){
    //If we are not unzipping nets, we could either split the registers in the bank or swap them
    generatePipelineAwareMovePipeRegSplitOrSwap(moveType, netlist, (registerBank**) block0, 
						(registerBank**) block1, numMove);
  }
  else{
    //We are able to swap, split or unzip registers
    generatePipelineAwareMovePipeRegSplitSwapOrUnzip(moveType, netlist, (registerBank**) block0, 
						     (registerBank**) block1, numMove);
  }
}

//We want to do something with two register banks, but we can't split them up.
//Therefore, we have to swap them.
void generatePipelineAwareMovePipeRegSwap(int *moveType, Netlist *netlist, 
					  registerBank** regBank0, registerBank** regBank1, int *numMove){
  //Make sure that the registers mapped to this block can fit into the other and vice verse.
  //Otherwise we should not perform the move.
  if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() ||
     (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay()){
    *moveType = MOVENULL;
  }
  else{
    //The difference between these two is really insignificant - we just keep them apart for potential
    // bookkeeping reasons.
    if((*regBank1)->isBankEmpty()){
      *moveType = MOVEREGEMPTYSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty register bank at delay block %d, %d bank %d to swap\n", (*regBank1)->getHostCLB()->get_col(), 
	     (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
#endif
    }
    else{
      *moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
      if((*regBank1)->get_currBlock()){
	printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap\n", 
	       (*regBank1)->get_currBlock()->name.c_str(), 
	       (*regBank1)->getHostCLB()->get_col(), 
	       (*regBank1)->getHostCLB()->get_row(),
	       (*regBank1)->getHostBankNum());	
      }
      else{
	printf("Selected register bank with mapped pipelining registers at delay block %d, %d bank to swap\n",
	       (*regBank1)->getHostCLB()->get_col(), 
	       (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
      }
#endif
    }
  }
}

//We want to do something with two register banks, but we can only split or swap, not duplicate
void generatePipelineAwareMovePipeRegSplitOrSwap(int *moveType, Netlist *netlist, registerBank **regBank0, 
						 registerBank **regBank1, int *numMove){
  long choice;

  //What we can do depends on the relationship between these two blocks.
  //We know that bank0 is not empty, but bank1 might be.
  if((*regBank1)->isBankEmpty()){
    // If we are allowed to split registers and bank1 is empty we could:
    //  1) Put all of the registers from bank0 into bank1 (only if they would fit, though)
    //  2) Split the registers and put some into bank1
    // We will assume we want to either swap or split registers with equal probability
    if((*regBank0)->getNumMappedRegs() <= (*regBank1)->getNumPossibleDelay()){
      choice = lrand48() % 2;
    }
    else{
      //If we cannot swap these blocks, we should try to split them
      choice = 1;
    }
    
    if(choice == 0){
      //Swap the full bank and the empty bank
      *moveType = MOVEREGEMPTYSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty register bank at delay block %d, %d bank %d to swap\n", (*regBank1)->getHostCLB()->get_col(), 
	     (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
#endif	
    }
    else{
      //Let's split this register
      generatePipelineAwareMoveSplitReg(moveType, netlist, regBank0, regBank1, numMove);
    }
  }
  else{
    // If bank1 is not empty  and we are allowed to split blocks, our
    //  choices depend on the relationship between the two 
    //  1) If one bank is the parent of the other and does not have fan-out AND the other bank is not mapped to
    //     a logic block, we could either:
    //     A) Swap their contents altogether
    //     B) Merge the registers 
    //     We will assume that if they are related we want to choose between swapping and merging with
    //         equal probability.
    //  2) Otherwise we have to swap their contents altogether
    //
    //See if the last register in one is parent of the first one in the other and does not have fan-out and 
    // is not mapped to a logic block
    if((!(*regBank0)->getLastSource()->hasFanOut() &&
	(*regBank0)->getLastSource() == (*regBank1)->getFirstSink()->sourcePipeliningNet->source) ||
       (!(*regBank1)->getLastSource()->hasFanOut() &&
	(*regBank1)->getLastSource() == (*regBank0)->getFirstSink()->sourcePipeliningNet->source) && 
       !(*regBank1)->get_currBlock()){
      // Should we swap or merge?  Well, first let's check to see if we could swap the contents
      //  at all
      if((*regBank0)->getNumMappedRegs() <= (*regBank1)->getNumPossibleDelay() &&
	 (*regBank1)->getNumMappedRegs() <= (*regBank0)->getNumPossibleDelay()){
	choice = lrand48() % 2;
      }
      else{
	//If we can't swap these blocks, we should try to merge them
	choice = 1;
      }
      
      if(choice == 0){
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected related register bank with mapped pipelining registers at delay block %d, %d bank %d to swap\n",
	       (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	       (*regBank1)->getHostBankNum());
#endif
      }
      else{
	//Let's try to merge these two blocks somehow
	generatePipelineAwareMoveMergeReg(moveType, netlist, regBank0, regBank1, numMove);
      }
    }
    else{
      //They are unrelated logic or one is mapped to a logic block so we can only try to swap them
      if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() ||
	 (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay()){
	// We can't swap because one is too large to fit into the other
	*moveType = MOVENULL;
      }
      else{
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	if((*regBank1)->get_currBlock()){
	  printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap\n", 
		 (*regBank1)->get_currBlock()->name.c_str(), 
		 (*regBank1)->getHostCLB()->get_col(), 
		 (*regBank1)->getHostCLB()->get_row(),
		 (*regBank1)->getHostBankNum());
	}
	else{
	  printf("Selected unrelated register bank with mapped pipelining registers at delay block %d, %d bank %d to swap\n",
		 (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
		 (*regBank1)->getHostBankNum());
	}
#endif
      }
    }
  }
}

//We have selected two register banks and we are allowed to split, swap or duplicate registers
void generatePipelineAwareMovePipeRegSplitSwapOrUnzip(int *moveType, Netlist *netlist, registerBank **regBank0, 
						      registerBank **regBank1, int *numMove){
  long choice;

  //What we can do depends on the relationship between these two blocks.
  //We know that bank0 is not empty, but bank1 might be.
  if((*regBank1)->isBankEmpty()){
    // If we are allowed to unzip nets, split registers and bank1 is empty 
    //  what we can do depends on the characteristics of this register bank.  If 
    //  the last register in this bank has a fanout greater than 1 we could 
    //  1) Put all of the registers from bank0 into bank1
    //  2) Split the registers and put some into bank1
    //  3) Unzip the nets in this block and duplicate all of the registers
    // If this register bank does not have a fanout greater than 1, then we should only try the
    //  first two since duplicating these registers doesn't make sense.
    if((*regBank0)->getLastSource()->hasFanOut()){
      // We will assume we want to either swap, split, or unzip registers with equal probability
      if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay()){
	//If the registers in bank 0 cannot fit into bank 1, we can only split them
	choice = 1;
      }
      else{
	//Otherwise we can do all three
	choice = lrand48() % 3;
      }
    }
    else{
      // We will assume we want to either swap or split registers with equal probability
      if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay()){
	//If the registers in bank 0 cannot fit into bank 1, we can only split them
	choice = 1;
      }
      else{
	//Otherwise we can swap or split them
	choice = lrand48() % 2;
      }
    }
    
    if(choice == 0){
      //Swap the full bank and the empty bank
      *moveType = MOVEREGEMPTYSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty delay block at %d, %d bank %d to swap\n", (*regBank1)->getHostCLB()->get_col(), 
	     (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
#endif	
    }
    else if(choice == 1){
      //Let's try to split this register
      generatePipelineAwareMoveSplitReg(moveType, netlist, regBank0, regBank1, numMove);
    }
    else{
      //Let's try to unzip these registers into the empty location
      *moveType = MOVEUNZIPPNET;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty delay block at %d, %d bank %d to unzip registers into\n", (*regBank1)->getHostCLB()->get_col(), 
	     (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
#endif	
      //generatePipelineAwareMoveUnzipReg(moveType, netlist, regBank0, regBank1, numMove);
    }
  }
  else{
    // If bank1 is not empty  and we are allowed to split and duplicate blocks, our
    //  choices depend on the relationship between the two 
    //  1) If one bank is the parent of the other and does not have fan-out AND the other bank is not mapped to
    //     a logic block, we could either:
    //     A) Swap their contents altogether
    //     B) Merge the registers 
    //     We will assume that if they are related we want to choose between swapping and merging with
    //         equal probability.
    //  2) If any of the registers in both banks have the same source net, we could either:
    //     A) Swap their contents altogether
    //     B) Zip the nets back together
    //  3) Otherwise we have to swap their contents altogether
    //
    //See if the last register in one is parent of the first one in the other and does not have fan-out and 
    // is not mapped to a logic block
    if((!(*regBank0)->getLastSource()->hasFanOut() &&
	(*regBank0)->getLastSource() == (*regBank1)->getFirstSink()->sourcePipeliningNet->source) ||
       (!(*regBank1)->getLastSource()->hasFanOut() &&
	(*regBank1)->getLastSource() == (*regBank0)->getFirstSink()->sourcePipeliningNet->source) &&
       !(*regBank1)->get_currBlock()){
      // Should we swap or merge?  Well, first let's check to see if we could swap the contents
      //  at all
      if((*regBank0)->getNumMappedRegs() <= (*regBank1)->getNumPossibleDelay() &&
	 (*regBank1)->getNumMappedRegs() <= (*regBank0)->getNumPossibleDelay()){
	choice = lrand48() % 2;
      }
      else{
	//If we can't swap these blocks, we should try to merge them
	choice = 1;
      }

      if(choice == 0){
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected related mapped delay block at %d, %d bank %d to swap\n", 
	       (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	       (*regBank1)->getHostBankNum());
#endif
      }
      else{
	//Let's try to merge these two blocks somehow
	generatePipelineAwareMoveMergeReg(moveType, netlist, regBank0, regBank1, numMove);
      }
    }
    //Check to see if any of the registers in one bank are duplicates of ones in the other bank.
    //We will only consider them to be duplicates if they both start from the same source net and end with the same
    // latency and the other bank is not mapped to a logic block.
    else if((*regBank0)->getFirstSink()->sourcePipeliningNet == (*regBank1)->getFirstSink()->sourcePipeliningNet &&
	    (*regBank0)->getNumMappedRegs() == (*regBank1)->getNumMappedRegs() && !(*regBank1)->get_currBlock()){
      // Should we swap or zip the nets back together?  
      // We don't have to check to see if we could swap the contents, since we already know they are duplicates
      //  of each other
      // Perhaps I should also add in the capability to swap sinks.  Since they are identical, that should be easy
      //  to do.
      choice = lrand48() % 2;

      if(choice == 0){
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected duplicate mapped delay block at %d, %d bank %d to swap\n", 
	       (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	       (*regBank1)->getHostBankNum());
#endif
      }
      else{
	*moveType = MOVEZIPPNET;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected duplicate delay block at %d, %d bank %d to zip\n", (*regBank1)->getHostCLB()->get_col(), 
	       (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
#endif	
	//generatePipelineAwareMoveZipReg(moveType, netlist, regBank0, regBank1, numMove);
      }
    }
    else{
      //They are unrelated logic so we have to swap
       if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() ||
	  (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay()){
	 // We can't swap because one is too large to fit into the other
	 *moveType = MOVENULL;
       }
       else{
	 *moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	if((*regBank1)->get_currBlock()){
	  printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap\n", 
		 (*regBank1)->get_currBlock()->name.c_str(), 
		 (*regBank1)->getHostCLB()->get_col(), 
		 (*regBank1)->getHostCLB()->get_row(),
		 (*regBank1)->getHostBankNum());
	}
	else{
	  printf("Selected unrelated register bank with mapped pipelining registers at delay block %d, %d bank %d to swap\n",
		 (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
		 (*regBank1)->getHostBankNum());
	}
#endif
       }
    }
  } 
}
  
//We want to split the registers in bank0 up and put some into the currently empty bank1
void generatePipelineAwareMoveSplitReg(int *moveType, Netlist *netlist, registerBank **regBank0, registerBank **regBank1, int *numMove){
  int tempInt;
  long choice;

  //Choose a random number of registers in this bank to bump out
  //Figure out how many registers we can pull out of block 0 and how many 
  // registers we can push into block1
  tempInt = min((*regBank0)->getNumMappedRegs(), (*regBank1)->getNumFreeRegs());	
	
  //Make sure this is some number between 1 and all of the registers
  *numMove = lrand48() % tempInt;	    
  *numMove = *numMove + 1;
  
  //This is just moving all of the registers from one to the other, 
  // so for correct bookkeeping I will catch this case.
  if(*numMove == (*regBank0)->getNumMappedRegs()){
    *moveType = MOVEREGEMPTYSWAP;
  }
  else{
    //We could either:
    //  A) Put the first N registers from block0, bank0 into block1, bank1
    //  B) Put the last N registers from block0, bank0 into block1, bank1
    //We'll assume to would like to choose between these with equal probability
    choice = lrand48() % 2;

    if(choice == 0){
      *moveType = MOVEREGEMPTYFIRST;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty delay block at %d, %d bank %d to move first %d register into\n", 
	     (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	     (*regBank1)->getHostBankNum(), *numMove);
#endif
    }
    else{
      *moveType = MOVEREGEMPTYLAST;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Selected empty delay block at %d, %d bank %d to move last %d register into\n", 
	     (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	     (*regBank1)->getHostBankNum(), *numMove);
#endif
    }
  }
}

//We want to merge the registers in bank 0 and bank 1
void generatePipelineAwareMoveMergeReg(int *moveType, Netlist *netlist, registerBank **regBank0, 
				       registerBank **regBank1, int *numMove){
  registerBank *tempBank;
  long choice;
  int numPossible;

  //First, let's try to set bank0 as the parent and bank1 as the child
  if((*regBank1)->getLastSource() == (*regBank0)->getFirstSink()->sourcePipeliningNet->source){
    tempBank = *regBank0;
    *regBank0 = *regBank1;
    *regBank1 = tempBank;
#if defined(PIPEPLACEDEBUGHEAVY)
    printf("Exchanged Block0 and Block1 to get proper parent child relationship\n");
#endif
  }

#if defined(PARANOID_CHECK_PIPEPLACE)
  //Verify that block 0 is the parent of block 1
  if((*regBank0)->getLastSource() != (*regBank1)->getFirstSink()->sourcePipeliningNet->source){
    ERROR("Trying to merge unrelated registers!");
  }

  //Verify that the last register in block 0 does not have any fan-out
  if((*regBank0)->getLastSource()->hasFanOut()){
    ERROR("Trying to merge a register that has fanout!");
  }
#endif

  // We could either 
  //  A) Pull the first N registers out of the child (block1) and put it into the parent (block0)
  //  B) Pull the last N registers out of the parent (block0) and put it into the child (block1)
  // Assume we want to choose between these possibilities evenly
  choice = lrand48() % 2;	

  if(choice == 0){
    //Figure out how many registers we can pull out of block 1 (child) and how many 
    // registers we can push into block0 (parent).
    numPossible = min((*regBank1)->getNumMappedRegs(), (*regBank0)->getNumFreeRegs());
    
    //The parent might already be full
    if(numPossible == 0){
      //See if we can swap these blocks
      if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() ||
	 (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay()){
	// We can't swap because one is too large to fit into the other
	*moveType = MOVENULL;
      }
      else{      
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected related delay block at %d, %d bank %d and delay block at %d, %d bank %d to swap because parent was full\n",
	       (*regBank0)->getHostCLB()->get_col(), (*regBank0)->getHostCLB()->get_row(), 
	       (*regBank0)->getHostBankNum(),
	       (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	       (*regBank1)->getHostBankNum());
#endif
      }
    }
    else{
      //Make sure this is some number between 1 and all of the registers we can transfer
      *numMove = lrand48() % numPossible;
      *numMove = *numMove + 1;		
      
      *moveType = MOVEREGPARENT;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Pulling first %d registers from child delay block at %d, %d bank %d and appending them to parent delay block at %d, %d bank %d\n",
	     *numMove, (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	     (*regBank1)->getHostBankNum(),
	     (*regBank0)->getHostCLB()->get_col(), (*regBank0)->getHostCLB()->get_row(), 
	     (*regBank0)->getHostBankNum());
#endif
    }
  }
  else{
    //Figure out how many registers we can pull out of block 0 (parent) and how many 
    // registers we can push into block1 (child).
    numPossible = min((*regBank0)->getNumMappedRegs(), (*regBank1)->getNumFreeRegs());

    //The child might already be full
    if(numPossible == 0){
      //See if we can swap these blocks
      if((*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() ||
	 (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay()){
	// We can't swap because one is too large to fit into the other
	*moveType = MOVENULL;
      }
      else{      
	*moveType = MOVEREGSWAP;
#if defined(PIPEPLACEDEBUGHEAVY)
	printf("Selected mapped delay block at %d, %d bank %d and delay block at %d, %d bank %d to swap because child was full\n",
	       (*regBank0)->getHostCLB()->get_col(), (*regBank0)->getHostCLB()->get_row(), 
	       (*regBank0)->getHostBankNum(),
	       (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	       (*regBank1)->getHostBankNum());
#endif
      }
    }
    else{
      //Make sure this is some number between 1 and all of the registers we can transfer
      *numMove = lrand48() % numPossible;
      *numMove = *numMove + 1;
      *moveType = MOVEREGCHILD;
#if defined(PIPEPLACEDEBUGHEAVY)
      printf("Pulling last %d registers from parent delay block at %d, %d bank %d and appending them to child delay block at %d, %d bank %d\n",
	     *numMove, (*regBank0)->getHostCLB()->get_col(), (*regBank0)->getHostCLB()->get_row(), 
	     (*regBank0)->getHostBankNum(),		   
	     (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	     (*regBank1)->getHostBankNum());
#endif
    }
  }
}

//Bank 0 has some registers that we would like to duplicate.
//Bank 1 should be empty
// void generatePipelineAwareMoveUnzipReg(int *moveType, Netlist *netlist, registerBank **regBank0, 
// 				       registerBank **regBank1, int *numMove){
//   //Because we assume that a register bank can only really have one output, if
//   // we are going to duplicate this bank, we have to duplicate the whole thing.
//   //Thus, we should check that block 1 can hold all of the register currently
//   // mapped to block 0
//   if((*regBank0)->getNumMappedRegs() <= (*regBank1)->getNumFreeRegs()){
//     *moveType = MOVEUNZIPPNET;
// #if defined(PIPEPLACEDEBUGHEAVY)
//     printf("Selected empty delay block at %d, %d bank %d to unzip registers into\n", 
// 	   (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
// 	   (*regBank1)->getHostBankNum());
// #endif	
//   }
//   else{
//     //If this register bank cannot be duplicated, they can't be swapped either.
//     //This means that they should be 
    
//   }
// }

// void generatePipelineAwareMoveZipReg(int *moveType, Netlist *netlist, registerBank **regBank0, 
// 				       registerBank **regBank1, int *numMove){

// }

//Check to see if it is possible to retime any registers backwards.
//Depending upon how things are connected we might need to grab new register banks
// and the architecture might not have any empty.
//If it is possible at all we will return true and how many we want to move.
//If not, we will return false.
bool generatePipelineAwareMoveRegRetimeBackward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove){
  logicBlock *lb;
  int index, i;
  list <int> numMoveList;
  list <int>::iterator intIter;
  int numRegsFreed, numRegsAdded;
  Terminal* tempTerm;

  //Grab the logic block that is currently mapped to this block
  if(blockType == LB){
    lb = ((vprLUT*)*block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)*block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to retime backwards!");
  }

  //Build a list of the number of register we can move back through this logic block - between 1 and all of the ones we can move
  for(i = 1; i <= lb->numRegCanPushBackward; i++){
    numMoveList.push_back(i);
  }

  while (numMoveList.size() != 0){
    //Grab a random number from the list, then delete it from the list
    index = lrand48() % numMoveList.size();
    for(intIter = numMoveList.begin(), i = 0; i < index; intIter++, i++){
    }
    *numMove = *intIter;
    numMoveList.erase(intIter);

    //Figure out how many registers this would free up on the output side.
    //Notice if a register bank is still partially occupied, this will not count them
    numRegsFreed = 0;

    //Go through all of the outputs and see what would happen if we remove this many
    // delays from each
    lb->create_outputListItr();
    for(tempTerm = lb->getnext_outputList(); tempTerm != 0; tempTerm = lb->getnext_outputList()){
      numRegsFreed += tempTerm->numRegsFreedForward(*numMove);
    }
    lb->remove_outputListItr();
    
#ifdef PLACER_DEBUG
    numRegistersEstimatedFreed = numRegsFreed;
#endif
    //cout << "Retiming " << *numMove << " registers backwards would free " << numRegsFreed << " registers " <<endl;

    //Figure out how many registers we would need to add on the input side
    //Notice we try to fill up existing, partially occupied banks if we can.
    numRegsAdded = 0;

    //Go through each of the inputs and figure out how many registers we could push on for free
    // (ones that happen to fill in empty spaces in banks that are already taken)
    // then figure out how many extra we'll need
    lb->create_inputListItr();
    for(tempTerm = lb->getnext_inputList(); tempTerm != 0; tempTerm = lb->getnext_inputList()){
      numRegsAdded += tempTerm->numRegsAddedBackward(*numMove);
    }
    lb->remove_inputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded = numRegsAdded;
#endif

    //cout << "Retiming " << *numMove << " register backwards would require " << numRegsAdded << " new registers " <<endl;

    //Do we have that many register banks free in the system?
    // If not, try some other number of registers
    if(numRegsFreed + FPGA->numFreeRegisters >= numRegsAdded){
        *moveType = MOVEREGRETIMEBACK;
	return true;
    }
  }

#ifdef PLACER_DEBUG
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  // We cannot move any number of registers back through this block
  return (false);  
}

//Check to see if it is possible to retime any registers forwards.
//Depending upon how things are connected we might need to grab new register banks
// and the architecture might not have any empty.
//If it is possible at all we will return true and how many we want to move.
//If not, we will return false.
bool generatePipelineAwareMoveRegRetimeForward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove){
  logicBlock *lb;
  int index, i;
  list <int> numMoveList;
  list <int>::iterator intIter;
  int numRegsFreed, numRegsAdded;
  Terminal* tempTerm;

  //Grab the logic block that is currently mapped to this block
  if(blockType == LB){
    lb = ((vprLUT*)*block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)*block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to retime backwards!");
  }

  //Build a list of the number of register we can move forward through this logic block - between 1 and all of the ones we can move
  for(i = 1; i <= lb->numRegCanPushForward; i++){
    numMoveList.push_back(i);
  }

  while (numMoveList.size() != 0){
    //Grab a random number from the list, then delete it from the list
    index = lrand48() % numMoveList.size();
    for(intIter = numMoveList.begin(), i = 0; i < index; intIter++, i++){
    }
    *numMove = *intIter;
    numMoveList.erase(intIter);

    //Figure out how many registers this would free up on the input side.
    //Notice if a register bank is still partially occupied, this will not count them
    numRegsFreed = 0;

    //Go through all of the inputs and see what would happen if we remove this many
    // delays from each
    lb->create_inputListItr();
    for(tempTerm = lb->getnext_inputList(); tempTerm != 0; tempTerm = lb->getnext_inputList()){
      //While we are doing this, figure out if the number of registers we want to
      // retime forwards is even available as an output.  That is,
      // we will only allow this function to retime to a currently
      // available signals if there is upstream fanout.  If we have a latency 9 
      ///signal in groups of 3 registers, we will only be able to retime backwards 
      // 3, 6 or 9 registers if we have seen any fanout in between.
      numRegsFreed += tempTerm->numRegsFreedBackward(*numMove);
    }
    lb->remove_inputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedFreed = numRegsFreed;
#endif

    //We will return a negative number if we can't retime this signal by this amount due to fan-out
    if(numRegsFreed < 0){
      continue;
    }

    //cout << "Retiming " << *numMove << " registers forwards would free " << numRegsFreed << " registers " <<endl;

    //Figure out how many registers we would need to add on the output side
    //Notice we try to fill up existing, partially occupied banks if we can.
    numRegsAdded = 0;

    //Go through each of the outputs and figure out how many registers we could push on for free
    // (ones that happen to fill in empty spaces in banks that are already taken)
    // then figure out how many extra we'll need
    lb->create_outputListItr();
    for(tempTerm = lb->getnext_outputList(); tempTerm != 0; tempTerm = lb->getnext_outputList()){
      numRegsAdded += tempTerm->numRegsAddedForward(*numMove);
    }
    lb->remove_outputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded = numRegsAdded;
#endif    

    //cout << "Retiming " << *numMove << " register forwards would require " << numRegsAdded << " new registers " <<endl;

    //Do we have that many register banks free in the system?
    // If not, try some other number of registers
    if(numRegsFreed + FPGA->numFreeRegisters >= numRegsAdded){
        *moveType = MOVEREGRETIMEFORWARD;
	return true;
    }
  }

  // We cannot move any number of registers back through this block
#ifdef PLACER_DEBUG
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  return (false);
}

//Commit a change to the system and calculate new delays and incremental slacks.
// Notice that these slack updates are not quite correct since the critical path might change
//We return true if the move was completed.
bool makePipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, 
			   enum e_timingMoveBlockType blockType, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;
  logicBlock* lb1;
  bool success;
  
  // We are trying to move a logic block or an IO pin
  if(*moveType == MOVELOGIC){
    // See if this is an I/O pin or a logic block
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);
      
      //Notice it doesn't matter if we are doing clb-level placement
      ((vprIOB*)(*block0))->swapMapping(index0, lb0, (vprIOB*)(*block1), index1, lb1);
    }
    else if(blockType == REGISTERBANK){
      if(!clb_level_placement){
	lb0 = ((registerBank*)(*block0))->get_currBlock();
	lb1 = ((registerBank*)(*block1))->get_currBlock();
	((registerBank*)(*block0))->swapMapping(lb0, (registerBank*)(*block1), lb1);
      }
      else{
	//((registerBank*)(*block0))->swapEntireCLBMappings(lb0, (registerBank*)(*block1), lb1);
	((registerBank*)(*block0))->swapEntireCLBMappings((registerBank*)(*block1));
      }
    }
    else if(blockType == LB){
      //Get the logic block associated with block 0 and 1
      if(!clb_level_placement){
	if(lutStickiness){
	  ((vprLUT*)(*block0))->swapStickyMapping((vprLUT*)(*block1));
	}
	else{
	  lb0=((vprLUT*)(*block0))->get_currBlock();
	  lb1=((vprLUT*)(*block1))->get_currBlock();      
	  ((vprLUT*)(*block0))->swapMapping(index0, lb0, (vprLUT*)(*block1), index1, lb1);
	}
      }
      else{
	//((vprLUT*)(*block0))->swapEntireCLBMappings(lb0, (vprLUT*)(*block1), lb1);
	((vprLUT*)(*block0))->swapEntireCLBMappings((vprLUT*)(*block1));
      }
    }
    else{
      ERROR("Unknown block type for move logic!");
    }
  }
  //We are moving a delay block bank into an empty delay block bank or swapping two delay block banks
  else if(*moveType == MOVEREGEMPTYSWAP || *moveType == MOVEREGSWAP){
    ((registerBank *)(*block0))->swapBank((registerBank *)(*block1));
   
    //Update the sink and slack timings for the nets we just moved
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
  }
  //We are moving the first N registers from block 0 to an empty block 1
  else if(*moveType == MOVEREGEMPTYFIRST || *moveType == MOVEREGEMPTYLAST ||
	  *moveType == MOVEREGPARENT || *moveType == MOVEREGCHILD){
    if(*moveType == MOVEREGEMPTYFIRST){
      ((registerBank *)(*block0))->migrateFirstNRegs((registerBank *)(*block1), numMove);
    }
    //We are moving the last N register from block 0 to an empty block 1
    else if(*moveType == MOVEREGEMPTYLAST){
      ((registerBank *)(*block0))->migrateLastNRegs((registerBank *)(*block1), numMove);
    }
    //We are moving the first N register from the child block 1 to the parent block 0
    else if(*moveType == MOVEREGPARENT){
      ((registerBank *)(*block1))->migrateFirstNRegs((registerBank *)(*block0), numMove);
    }
    //We are moving the last N register from the parent block 0 to the child block 1
    else if(*moveType == MOVEREGCHILD){
      ((registerBank *)(*block0))->migrateLastNRegs((registerBank *)(*block1), numMove);
    }

    //Update the sink and slack timings for the nets we just moved
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
  }
  else if(*moveType == MOVEUNZIPPNET){
    ((registerBank *)(*block0))->duplicateRegs(netlist, (registerBank *)(*block1), criticalPathDelay);

    //Update the sink and slack timings for the nets we just modified
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
  }
  else if(*moveType == MOVEZIPPNET){
    ((registerBank *)(*block0))->zipRegs(netlist, (registerBank *)(*block1), criticalPathDelay);

    //Update the sink and slack timings for the nets we just modified.
    //Block 1 will be empty at this time.
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
  }
  else if(*moveType == MOVEREDOUNZIPPNET){
    ((registerBank *)(*block0))->reDuplicateRegs(netlist, (registerBank *)(*block1), criticalPathDelay);
  
    //Update the sink and slack timings for the nets we just modified
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
  }
  //We want to move N registers from each output of block 0 to the input of block 0
  else if(*moveType == MOVEREGRETIMEBACK){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      if(!((vprLUT*)(*block0))->get_currBlock()->retimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed backwards.
	//Undo the partially completed retiming.
	//netlist->printPlacement();
	//netlist->print_netList();
	
	((vprLUT*)(*block0))->get_currBlock()->abortRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did a regular undo retime backward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	
	*moveType = MOVEREGRETIMEBACKABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We succeeded in the retiming, so update all of the sinks and slacks for the
	//new nets we created upstream
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
      }
    }
    else if(blockType == REGISTERBANK){
      if(!((registerBank*)(*block0))->get_currBlock()->retimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed backwards.
	//Undo the partially completed retiming.
	//netlist->printPlacement();
	//netlist->print_netList();
	
	((registerBank*)(*block0))->get_currBlock()->abortRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did a regular undo retime backward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	
	*moveType = MOVEREGRETIMEBACKABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We succeeded in the retiming, so update all of the sinks and slacks for the
	//new nets we created upstream
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
      }
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGRETIMEFORWARD){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      if(!((vprLUT*)(*block0))->get_currBlock()->retimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed forwards.
	//Undo the partially completed retiming.
	((vprLUT*)(*block0))->get_currBlock()->abortRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did an undo retime forward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	
	*moveType = MOVEREGRETIMEFORWARDABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We successfully added registers downstream of this logic block, so update the
	// new sink delays and slacks!
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
      }
    }
    else if(blockType == REGISTERBANK){
      if(!((registerBank*)(*block0))->get_currBlock()->retimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed forwards.
	//Undo the partially completed retiming.
	((registerBank*)(*block0))->get_currBlock()->abortRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did an undo retime forward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	
	*moveType = MOVEREGRETIMEFORWARDABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We successfully added registers downstream of this logic block, so update the
	// new sink delays and slacks!
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
      }
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGUNDOREGRETIMEBACK){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      ((vprLUT*)(*block0))->get_currBlock()->undoRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->get_currBlock()->undoRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGUNDOREGRETIMEFORWARD){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      ((vprLUT*)(*block0))->get_currBlock()->undoRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->get_currBlock()->undoRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVENULL){

  }
  else{
    ERROR("Unknown move type!");
  }
  return true;
}

//We made a change to the system and we would like to undo it.
void undoPipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, enum e_timingMoveBlockType blockType, int numMove,
			     double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  
  // We just did a normal swap operation last time, so undoing it is easy
  // Since before we swapped block 1 might have been empty, now block 0 might be empty, so we put it
  //  as the second argument
  if(*moveType == MOVELOGIC || *moveType == MOVEREGEMPTYSWAP || *moveType == MOVEREGSWAP){
    makePipelineAwareMove(moveType, netlist, block1, block0, index1, index0, blockType, numMove, criticalPathDelay, NULL);
  }

  //If we split the first N from block 0 into an empty block 1, we should reverse the operation
  // by appending the registers from the parent (block1) back to the child (block0)
  else if(*moveType == MOVEREGEMPTYFIRST){
    *moveType = MOVEREGCHILD;
    makePipelineAwareMove(moveType, netlist, block1, block0, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we split the last N from block 0 into an empty block 1, we should reverse the operation
  // by appending the registers from the child (block1) back into the parent (block0)
  else if(*moveType == MOVEREGEMPTYLAST){
    *moveType = MOVEREGPARENT;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //We moved the first N registers from the child (block1) to the parent (block0) so to reverse the operation
  // we should move the last N registers from the parent to the child
  else if(*moveType == MOVEREGPARENT){
    *moveType = MOVEREGCHILD;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //We moved the last N registers from the parent (block0) to the child (block1) so to reverse the operation
  // we should move the first N registers from the child to the parent
  else if(*moveType == MOVEREGCHILD){
    *moveType = MOVEREGPARENT;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we just duplicated the registers from block 0 into block 1, to reverse this, we just zip them back into block 0
  else if(*moveType == MOVEUNZIPPNET){
    *moveType = MOVEZIPPNET;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we just zipped the registers from block 1 back into block 0, we can reverse this, but we have to make sure to
  // re-connect the correct sinks (still held by the registers and pnets sitting in block1) back to block 1 for this move to 
  // be accurately reversed.
  else if(*moveType == MOVEZIPPNET){
    *moveType = MOVEREDOUNZIPPNET;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  else if(*moveType == MOVEREGRETIMEBACK){
    *moveType = MOVEREGUNDOREGRETIMEBACK;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  retimeUndoList);
  }
  else if(*moveType == MOVEREGRETIMEFORWARD){
    *moveType = MOVEREGUNDOREGRETIMEFORWARD;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  retimeUndoList);
  }
  else{
    ERROR("Unknown move type!");
  }
}

//Evaluate the cost of the given move, but use the "delta delay" method.
//We usually do:
//  timing cost of a source-sink pair = point to point delay * criticality of source-sink pair ^ exp
//  deltaTimingCost = (new delay * new crit)^exp - (old delay * old crit)^exp
//  deltaCost = (alpha * ALL NETS deltaWireCost / ALL NETS oldWireCost) + 
//               [(1-alpha) * ALL NETS deltaTimingCost / ALL NETS oldTimingCost]
//  ALL NETS newTimingCost = ALL NETS oldTimingCost + ALL NETS deltaTimingCost
//
//However, to avoid issues where the delta cost of a net gets weird when we
// update the slack incrementally each time we move a block, we would like to calculate the timing cost
// and delta cost a bit differently.
//
//First, the timing cost is calculated almost the same:
//  timing cost of a source-sink pair = point to point delay * criticality of source-sink pair
//that is, the timing cost is still dependant on the old vs. new criticality (what was potentially making the
//  delta cost weird earlier), but taking out the exponent.
//Thus, the delta timing cost is still basically the same:
// deltaTimingCost = (new delay * new crit) - (old delay * old crit)
//
//The big difference that we make is that we create a second "delta timing cost" equation:
// deltaTimingCost' = (new delay - old delay) * new crit^exp
//
//Then we use this delta timing cost' in our overall delta cost function:
// deltaCost = (alpha * ALL NETS deltaWireCost / ALL NETS oldWireCost) + 
//               [(1-alpha) * ALL NETS deltaTimingCost' / ALL NETS oldTimingCost]
//
//However, the newTimingCost is still calculated the same
// ALL NETS newTimingCost = ALL NETS oldTimingCost + ALL NETS deltaTimingCost
//
double determineDeltaCostDeltaDelay(Netlist *netlist, int moveType, void** block0, void** block1, int index0, int index1, 
			  enum e_timingMoveBlockType blockType, int numMove,
				    bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  double deltaWire = 0.0;
  double deltaTiming = 0.0;
  double deltaTimingPrime = 0.0;
  double deltaCost = 0.0;
  double newMaxCut = 0.0;
  double deltaMaxCut = 0.0;
  double newAveCut = 0.0;
  double deltaAveCut = 0.0;

  //Go through the nets associated with the two blocks and record the change in their wirelength and timing.
  //We have to make sure that we do not double-count this change if they are connected to each other!
  if(moveType == MOVELOGIC){
    // See if this is an I/O pin or a logic block
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);

      //Since we swapped them, block 0 might be the empty one now!
      //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
      lb1->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);
      if(lb0)
	lb0->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);      
    }
    else if(blockType == REGISTERBANK){
      if(!clb_level_placement){
	//Get the logic block associated with block 0 and 1
	lb0=((registerBank*)(*block0))->get_currBlock();
	lb1=((registerBank*)(*block1))->get_currBlock();

	//Since we swapped them, block 0 might be the empty one now!
	//Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
	lb1->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);
	if(lb0)
	  lb0->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);  

	// If we did a swap between a forced explicit register and
	// a group of pipelining registers this won't be good enough
	// since there is no logic block associated with pipelining registers.
	//However we can tell this apart from simply swapping with an empty location
	// since get_currBlock() will come up NULL but isBankEmpty will not.
	if(!lb0 && !((registerBank*)(*block0))->isBankEmpty()){
	  ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
	  
	  //Figure out what the overall change in wirelength and timing was
	  deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
	  
	  if(timing_driven){
	    deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
	    deltaTimingPrime += ((registerBank *)(*block0))->determineDeltaTimingCostPrime();
	  }	
	}
      }
      else{
	((registerBank*)(*block0))->updateCostAndDeltasForEntireCLBCostDeltaDelay((registerBank*)(*block1), timing_driven, criticalPathDelay, 
									      &deltaWire, &deltaTiming, &deltaTimingPrime);
      }
    }
    else if(blockType == LB){
      if(!clb_level_placement){
	if(lutStickiness){
	  if(lut_level_placement){
	    ERROR("LUT-level placement with sticky LUTs has not been implemented yet!");
	  }
	  else{
	    //Update the costs and deltas for the LUTs we moved
	    ((vprLUT*)(*block0))->updateCostAndDeltasForAllCLBLUTsCostDeltaDelay((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
									     &deltaWire, &deltaTiming, &deltaTimingPrime);
	  
	    //Update the costs and deltas for all the FFs we moved.
	    FPGA->getCLBAt(((vprLUT*)(*block0))->get_row(), ((vprLUT*)(*block0))->get_col())->updateCostAndDeltasForAllFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
																     &deltaWire, &deltaTiming, &deltaTimingPrime);
	    FPGA->getCLBAt(((vprLUT*)(*block1))->get_row(), ((vprLUT*)(*block1))->get_col())->updateCostAndDeltasForAllFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
																     &deltaWire, &deltaTiming, &deltaTimingPrime);
	  }
	}
	else{
	  if(lut_level_placement){
	    //Get the logic block associated with block 0 and 1
	    lb0=((vprLUT*)(*block0))->get_currBlock();
	    lb1=((vprLUT*)(*block1))->get_currBlock();      
	    
	    //Since we swapped them, block 0 might be the empty one now!
	    //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
	    lb1->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);
	    if(lb0)
	      lb0->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);
	  }
	  else{
	    ((vprLUT*)(*block0))->updateCostAndDeltasForAllCLBLUTsCostDeltaDelay((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
										 &deltaWire, &deltaTiming, &deltaTimingPrime);
	  } 
	}
      }
      else{
	((vprLUT*)(*block0))->updateCostAndDeltasForEntireCLBCostDeltaDelay((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
									    &deltaWire, &deltaTiming, &deltaTimingPrime);
      }
    }
    else{
      ERROR("Unknown move block type!");
    }
  }

  //We are moving a delay block bank into an empty delay block bank or swapping two delay block banks
  else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP || 
	  moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST ||
	  moveType == MOVEREGPARENT || moveType == MOVEREGCHILD ||
	  moveType == MOVEUNZIPPNET || moveType == MOVEREDOUNZIPPNET){
    //Update the wire and timing costs for the nets we just moved based on where they moved to and the changes made to the slack vectors
    ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    ((registerBank *)(*block1))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    
    //Figure out what the overall change in wirelength and timing was
    deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
    //cout << "Delta Wire after block 0 = " << deltaWire << endl;
    deltaWire += ((registerBank *)(*block1))->determineDeltaWireCost();
    //cout << "Delta Wire after block 1 = " << deltaWire << endl;
    if(timing_driven){
      deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 0 = " << deltaTiming << endl;
      deltaTimingPrime += ((registerBank *)(*block0))->determineDeltaTimingCostPrime();

      deltaTiming += ((registerBank *)(*block1))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 1 = " << deltaTiming << endl;
      deltaTimingPrime += ((registerBank *)(*block1))->determineDeltaTimingCostPrime();
    }
  }

  else{
    ERROR("Delta delay moves not compatible with zipping or retiming moves");
  }

  if(timing_driven){
    //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
    deltaCost = tdPlacementLambda * (deltaTimingPrime / netlist->getTimingCost()) + ((1.0 - tdPlacementLambda) * (deltaWire / netlist->getWireCost()));

    netlist->setWireCost(deltaWire + netlist->getWireCost());
    netlist->setTimingCost(deltaTiming + netlist->getTimingCost());

    return(deltaCost);
  }
  else{
    ERROR("Delta delays moves should not be used with congestion-driven placement!");
  }
}

//We've made a given move, now we just have to compute the delta cost
double determineDeltaCost(Netlist *netlist, int moveType, void** block0, void** block1, 
			  int index0, int index1, enum e_timingMoveBlockType blockType, int numMove,
			  bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  double deltaWire = 0.0;
  double deltaTiming = 0.0;
  double deltaCost = 0.0;
  double newMaxCut = 0.0;
  double deltaMaxCut = 0.0;
  double newAveCut = 0.0;
  double deltaAveCut = 0.0;

  //Go through the nets associated with the two blocks and record the change in their wirelength and timing.
  //We have to make sure that we do not double-count this change if they are connected to each other!
  if(moveType == MOVELOGIC){
    // See if this is an I/O pin or a logic block
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);

      //Since we swapped them, block 0 might be the empty one now!
      //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
      lb1->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
      if(lb0)
	lb0->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);      
    }
    else if(blockType == REGISTERBANK){
      if(!clb_level_placement){
	//Get the logic block associated with block 0 and 1
	lb0=((registerBank*)(*block0))->get_currBlock();
	lb1=((registerBank*)(*block1))->get_currBlock();

	//Since we swapped them, block 0 might be the empty one now!
	//Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
	lb1->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	if(lb0)
	  lb0->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);  

	// If we did a swap between a forced explicit register and
	// a group of pipelining registers this won't be good enough
	// since there is no logic block associated with pipelining registers.
	//However we can tell this apart from simply swapping with an empty location
	// since get_currBlock() will come up NULL but isBankEmpty will not.
	if(!lb0 && !((registerBank*)(*block0))->isBankEmpty()){
	  ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
	  
	  //Figure out what the overall change in wirelength and timing was
	  deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
	  //cout << "Delta Wire after block 0 = " << deltaWire << endl;
	  
	  if(timing_driven){
	    deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
	    //cout << "Delta Timing after block 0 = " << deltaTiming << endl;
	  }	
	}
      }
      else{
	//((registerBank*)(*block0))->updateCostAndDeltasForEntireCLB(lb0, lb1, (registerBank*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	((registerBank*)(*block0))->updateCostAndDeltasForEntireCLB((registerBank*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
      }
    }
    else if(blockType == LB){
      if(!clb_level_placement){
	if(lutStickiness){
	  if(lut_level_placement){
	    ERROR("LUT-level placement with sticky LUTs has not been implemented yet!");
	  }
	  else{
	    //Update the costs and deltas for the LUTs we moved
	    ((vprLUT*)(*block0))->updateCostAndDeltasForAllCLBLUTs((vprLUT*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	  
	    //Update the costs and deltas for all the FFs we moved.
	    FPGA->getCLBAt(((vprLUT*)(*block0))->get_row(), ((vprLUT*)(*block0))->get_col())->updateCostAndDeltasForAllFFs(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	    FPGA->getCLBAt(((vprLUT*)(*block1))->get_row(), ((vprLUT*)(*block1))->get_col())->updateCostAndDeltasForAllFFs(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	  }
	}
	else{
	  if(lut_level_placement){
	    //Get the logic block associated with block 0 and 1
	    lb0=((vprLUT*)(*block0))->get_currBlock();
	    lb1=((vprLUT*)(*block1))->get_currBlock();      
	    
	    //Since we swapped them, block 0 might be the empty one now!
	    //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
	    lb1->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	    if(lb0)
	      lb0->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	  }
	  else{
	    //((vprLUT*)(*block0))->updateCostAndDeltasForAllCLBLUTs(lb0, lb1, (vprLUT*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	    ((vprLUT*)(*block0))->updateCostAndDeltasForAllCLBLUTs((vprLUT*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	  } 
	}
      }
      else{
	//((vprLUT*)(*block0))->updateCostAndDeltasForEntireCLB(lb0, lb1, (vprLUT*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
	((vprLUT*)(*block0))->updateCostAndDeltasForEntireCLB((vprLUT*)(*block1), timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
      }
    }
    else{
      ERROR("Unknown move block type!");
    }
  }

  //We are moving a delay block bank into an empty delay block bank or swapping two delay block banks
  else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP || 
	  moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST ||
	  moveType == MOVEREGPARENT || moveType == MOVEREGCHILD ||
	  moveType == MOVEUNZIPPNET || moveType == MOVEREDOUNZIPPNET){
    //Update the wire and timing costs for the nets we just moved based on where they moved to and the changes made to the slack vectors
    ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    ((registerBank *)(*block1))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    
    //Figure out what the overall change in wirelength and timing was
    deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
    //cout << "Delta Wire after block 0 = " << deltaWire << endl;
    deltaWire += ((registerBank *)(*block1))->determineDeltaWireCost();
    //cout << "Delta Wire after block 1 = " << deltaWire << endl;
    if(timing_driven){
      deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 0 = " << deltaTiming << endl;
      deltaTiming += ((registerBank *)(*block1))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 1 = " << deltaTiming << endl;
    }
  }

  else if(moveType == MOVEZIPPNET){
    ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);

    //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){
      ((registerBank *)(*block1))->updateDeletedCutCost();
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
    //cout << "Delta Wire after block 0 = " << deltaWire << endl;

    //We have to remove the cost of the nets we don't need anymore - this is all the nets in block 1
    deltaWire -= ((registerBank *)(*block1))->returnTotalWireCost();
    //cout << "Delta Wire after block 1 = " << deltaWire << endl;

    if(timing_driven){
      deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 0 = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets in block 1
      deltaTiming -= ((registerBank *)(*block1))->returnTotalTimingCost();
      //cout << "Delta Timing after block 1 = " << deltaTiming << endl;
    }
  }

  else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGUNDOREGRETIMEFORWARD){
    if(moveType == MOVEREGUNDOREGRETIMEFORWARD){
      if(blockType == LB){
	lb0=((vprLUT*)(*block1))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block1))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }
    else{
      if(blockType == LB){
	lb0=((vprLUT*)(*block0))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block0))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsBackwards(timing_driven, criticalPathDelay);

   //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){      
      updateDeletedCutCostForwards(retimeUndoList);
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += lb0->determineDeltaWireCostBackwards();
    //cout << "Delta Wire after logic block = " << deltaWire << endl;
    
    //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
    deltaWire -= returnTotalWireCostForwards(retimeUndoList);
    //cout << "Delta Wire after undo list = " << deltaWire << endl;
    
    //We might have created pipelined nets for non-pipelined nets.  Take account of these.
    deltaWire -= netlist->wireCostSurplus;
    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      deltaTiming += lb0->determineDeltaTimingCostBackwards();
      //cout << "Delta Timing after logic block = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
      deltaTiming -= returnTotalTimingCostForwards(retimeUndoList);
      //cout << "Delta Timing after undo list = " << deltaTiming << endl;

      deltaTiming -= netlist->timingCostSurplus;
      netlist->timingCostSurplus = 0.0;
    }
  }
  else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGUNDOREGRETIMEBACK){
    if(moveType == MOVEREGUNDOREGRETIMEBACK){
      if(blockType == LB){
	lb0=((vprLUT*)(*block1))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block1))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }
    else{
      if(blockType == LB){
	lb0=((vprLUT*)(*block0))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block0))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsForwards(timing_driven, criticalPathDelay);

    //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){      
      updateDeletedCutCostBackwards(retimeUndoList, lb0);
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += lb0->determineDeltaWireCostForwards();
    //cout << "Delta Wire after logic block = " << deltaWire << endl;
    
    // We have to remove the cost of the nets we don't need anymore 
    deltaWire -= returnTotalWireCostBackwards(retimeUndoList, timing_driven, criticalPathDelay, lb0);
    //cout << "Delta Wire after undo list = " << deltaWire << endl;

    //We might have had orphaned nets due to fan-out/undoing retiming or we might have created
    // pipelined nets for non-pipelined nets.  Take account of these.
    deltaWire -= netlist->wireCostSurplus;
    netlist->wireCostSurplus = 0.0;
    //cout << "Delta Wire after surplus = " << deltaWire << endl;
    
    if(timing_driven){
      deltaTiming += lb0->determineDeltaTimingCostForwards();
      //cout << "Delta Timing after logic block = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
      deltaTiming -= returnTotalTimingCostBackwards(retimeUndoList, lb0);
      //cout << "Delta Timing after undo list = " << deltaTiming << endl;

      deltaTiming -= netlist->timingCostSurplus;
      netlist->timingCostSurplus = 0.0;
      //cout << "Delta Timing after surplus = " << deltaTiming << endl;
    }
  }

  else if(moveType == MOVEREGRETIMEBACKABORT){
    if(blockType == LB){
      lb0=((vprLUT*)(*block0))->get_currBlock();
    }
    else if(blockType == REGISTERBANK){
      lb0=((registerBank*)(*block0))->get_currBlock();
    }
    else{
      ERROR("Block type not allowed to be retimed!");
    }    

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsForwards(timing_driven, criticalPathDelay);

    //Figure out what the overall change in wirelength and timing was,
    // but throw it out.
    lb0->determineDeltaWireCostForwards();
    
    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      lb0->determineDeltaTimingCostForwards();
      netlist->timingCostSurplus = 0.0;
    }
  }

  else if(moveType == MOVEREGRETIMEFORWARDABORT){
    if(blockType == LB){
      lb0=((vprLUT*)(*block0))->get_currBlock();
    }
    else if(blockType == REGISTERBANK){
      lb0=((registerBank*)(*block0))->get_currBlock();
    }
    else{
      ERROR("Block type not allowed to be retimed!");
    }    
    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsBackwards(timing_driven, criticalPathDelay);

    //Figure out what the overall change in wirelength and timing was,
    // but throw it out.
    lb0->determineDeltaWireCostBackwards();

    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      lb0->determineDeltaTimingCostBackwards();
      netlist->timingCostSurplus = 0.0;
    }
  }

  else{
    ERROR("Unknown move type");
  }

  //Figure out the new max cut  
  if(maxCutCost){
    //Go through the cut cost vector to determine the largest value
    for(vector <double>::iterator iter = FPGA->cutCostVector.begin(); iter != FPGA->cutCostVector.end(); iter++){
      //printf("Cut Cost = %d\n", (int) *iter);
      if(aveCutCost){
	newAveCut += *iter;
      }

      if(newMaxCut < *iter){
	//printf("New Max Cut = %f\n", *iter);
	newMaxCut = *iter;
      }
    }
    deltaMaxCut = newMaxCut - netlist->maxCut;

    if(aveCutCost){
      newAveCut /= FPGA->cutCostVector.size();
    }
    deltaAveCut = newAveCut - netlist->aveCut;
  } 

  //cout << "Delta Wire = " << deltaWire << endl;
  //cout << "Delta Timing = " << deltaTiming << endl;
  //if(maxCutCost){
    //cout << "Delta MaxCut = " << deltaMaxCut << endl;
  //}
  //if(aveCutCost){
    //cout << "Delta AveCut = " << deltaAveCut << endl;
  //}

  if(timing_driven){
    if(maxCutCost && !aveCutCost){
      //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
      //However, if we are paying attention to max cut, we have to incorporate that too.
      deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + 
	((1.0 - tdPlacementLambda) * (0.7*(deltaWire / netlist->getWireCost()) + 0.3 * (deltaMaxCut/ netlist->maxCut)));
      netlist->maxCut = newMaxCut;
    }
    else if(maxCutCost && aveCutCost){
      deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + 
	((1.0 - tdPlacementLambda) * (0.7*(deltaAveCut / netlist->aveCut) + 0.3 * (deltaMaxCut/ netlist->maxCut)));
      netlist->maxCut = newMaxCut;
      netlist->aveCut = newAveCut;
    }
    else{
      if(deltaDelayCost){

      }
      else{
	//If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
	deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + ((1.0 - tdPlacementLambda) * (deltaWire / netlist->getWireCost()));
      }
    }
    netlist->setWireCost(deltaWire + netlist->getWireCost());
    netlist->setTimingCost(deltaTiming + netlist->getTimingCost());
    //cout << "Delta Wire Cost = " << deltaWire << endl;
    //cout << "Wire Cost = " << netlist->getWireCost() << endl;
    //cout << "Delta Timing Cost = " << deltaTiming << endl;
    //cout << "Timing Cost = " << netlist->getTimingCost() << endl;
    //cout << "DeltaCost = " << deltaCost << endl << endl <<endl;

    return(deltaCost);
  }
  else{
    if(maxCutCost && !aveCutCost){
      //If we are not timing-driven, only consider the max cut and wirelength impact
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      netlist->maxCut = newMaxCut;

      deltaCost = (0.7 * deltaWire) + (0.3 * deltaMaxCut);
      return deltaCost;
    }
    else if(maxCutCost && aveCutCost){
      //If we are not timing-driven, only consider the max cut and ave cut
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      netlist->maxCut = newMaxCut;
      netlist->aveCut = newAveCut;
      
      deltaCost = (0.7 * deltaAveCut) + (0.3 * deltaMaxCut);
      return deltaCost;
    }
    else{
      //If we are not timing-driven, only consider the wirelength impact
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      //cout << "DeltaCost = " << deltaWire << endl;
      return deltaWire;
    }
  }
}

//We removed all of these nets up to the nerfed registers (temporarily) 
// so decrement the cut costs associated with these nets
void updateDeletedCutCostForwards(list <PipeliningNet *> *retimeUndoList){
  //Decrement the span of this pipelining net
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    (*pnIter)->decrementNerfedSpanForwards();
  }
}

//We removed all of these nets up to the nerfed registers (temporarily) 
// so decrement the cut costs associated with these nets
void updateDeletedCutCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list<Terminal*>::iterator sinkIter;

  //Decrement the span of this pipelining net
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    (*pnIter)->decrementNerfedSpanBackwards(*sinkIter);
  }
}

//We removed all of these nets up to the nerfed registers (temporarily)
// so we need to know what their wire cost was
double returnTotalWireCostForwards(list <PipeliningNet *> *retimeUndoList){
  double wireCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    wireCost += (*pnIter)->returnNerfedWireCostForwards();
  }
  return wireCost;
}

//We removed all of these nets up to the nerfed registers (temporarily)
// so we need to know what their wire cost was.
//Since we might have orphaned a netlist, this will also include an evaluation of the last, unnerfed net.
//Since this actually does the reCalculatePlacementCost, it must be used before the returnNerfedTimingCostBackwards function.
double returnTotalWireCostBackwards(list <PipeliningNet *> *retimeUndoList, bool timing_driven, double criticalPathDelay, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list<Terminal*>::iterator sinkIter;
  double wireCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    wireCost += (*pnIter)->returnNerfedWireCostBackwards(timing_driven, criticalPathDelay, *sinkIter);
  }
  return wireCost;
}

double returnTotalTimingCostForwards(list <PipeliningNet *> *retimeUndoList){
  double timingCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    timingCost += (*pnIter)->returnNerfedTimingCostForwards();
  }
  return timingCost;
}

double returnTotalTimingCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list <Terminal *>:: iterator sinkIter;
  double timingCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    timingCost += (*pnIter)->returnNerfedTimingCostBackwards(*sinkIter);
  }
  return timingCost;
}

//We have decided to accept the retiming move, so get rid of the nets
// and nerfed registers.
void clearRetimeUndoListForward(list <PipeliningNet *> *retimeUndoList){
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and start deleting nets and registers
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    (*pnIter)->deleteNerfedNetsAndRegistersForwards();
    //cout << "Freeing net " << (*pnIter)->name << endl;
    delete (*pnIter);
  }

  retimeUndoList->clear();
}

//We have decided to accept the retiming move, so get rid of the nets
// and nerfed registers.
void clearRetimeUndoListBackward(list <PipeliningNet *> *retimeUndoList, void *block, enum e_timingMoveBlockType blockType){
  logicBlock *lb;
  list <PipeliningNet *>::iterator pnIter;
  list <Terminal *>:: iterator sinkIter;

  if(blockType == LB){
    lb = ((vprLUT*)block)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)block)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }

  //Go through each of these pipelining nets, up to the nerfed registers,
  // and start deleting nets and registers.
  //Notice that there may be fewer nets in the retime undo list than there are
  // input pins.  That's because we might have aborted a retiming.
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    (*pnIter)->deleteNerfedNetsAndRegistersBackwards(*sinkIter);
  }

  retimeUndoList->clear();
}

//We have decided to accept the retiming move, so un-nerf the registers
// upstream of this logic block
void clearNerfedBlocksBackward(void* block0, enum e_timingMoveBlockType blockType){
  logicBlock *lb;

  if(blockType == LB){
    lb=((vprLUT*)block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb=((registerBank*)block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }
  lb->clearNerfedBlocksBackward();
}

//We have decided to accept the retiming move, so un-nerf the registers
// downstream of this logic block
void clearNerfedBlocksForward(void* block0, enum e_timingMoveBlockType blockType){
  logicBlock *lb;
  if(blockType == LB){
    lb=((vprLUT*)block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb=((registerBank*)block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }
  lb->clearNerfedBlocksForward();
}

// Anneal the netlist
void pipelineAnneal(Netlist *netlist, bool timing_driven, double initialTemperature, float startRLim){
  void* block0;
  void* block1;
  int indexOrBank0, indexOrBank1;
  int moveType;
  enum e_timingMoveBlockType blockType;
  int numMove;
  double dCost;
  float rLim = startRLim;
  int movesRejected;
  int nonZeroMovesAccepted;
  double criticalPathDelay;
  double deltaCostAccept;
  double deltaCostReject;
  double averageDeltaCostAccept;
  double averageDeltaCostNonZeroAccept;
  double averageDeltaCostReject;
  int numMoveLogic;
  int numRegSwap;
  int numRegSplit;
  int numRegMerge;
  int numZip;
  int numUnzip;
  int numMoveLogicReject;
  int numRegSwapReject;
  int numRegSplitReject;
  int numRegMergeReject;
  int numZipReject;
  int numUnzipReject;
  double aveRegSplit;
  double aveRegMerge;

  int numMoveRegRetime;
  double aveRegRetime;
  int numRegRetimeReject;
  double aveCostRegRetime;

  list <PipeliningNet *> retimeUndoList;

  bool moveSucceeded;
  int numMovesFailed;

  double averageOverallCost;
  double stdDevOverallCost;

#if defined(PARANOID_CHECK_PIPEPLACE)
  double oldWireCost;
  double oldTimingCost;
  double oldMaxCut;
  double oldAveCut;
  double reallyOldWireCost;
  double reallyOldTimingCost;
  double reallyOldMaxCut;
  double reallyOldAveCut;  
#endif
  
  double currTemp = initialTemperature;
  bool done = false;

#if defined (PIPEPLACEDEBUGHEAVY)
  int numberMoves = 0;
#endif

  // The number of moves per temperature is # moveable blocks ^ (4/3)
  int numMovesPerTemp = (int)(placeInnerLoop * pow((double)netlist->calculateNumBlocksAndPipeliningRegisters(), 1.33));
  int numMovesBeforeUpdate = (int)(tdPlacementUpdateInterval * (double) numMovesPerTemp);

  if(numMovesBeforeUpdate <= 0){
    numMovesBeforeUpdate = 1;
  }


  // Calculate everything from scratch before we begin
  criticalPathDelay =  calculateCriticalPathDelay(netlist, true);
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  printf("Moves per Temperature iteration = %d\n", numMovesPerTemp);

#if defined(PIPEPLACEDEBUG)
  if(timing_driven){
    if(!splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCPDelay\tWCost\t\tAvgOverallCost\tTimingCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\n");
      printf("--------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tAvgOverallCost\tTimingCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tAveRSp\tAveRMer\tSpMerR\n");
      printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && unzipNets){
//       printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\tAveRSp\tAveRMer\n");
//       printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\n");
      printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
  }
  else{
    if(!splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost");
      printf("--------------------------------------------------------------------------------------------------\n");    
    }
    else if(splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tAveRSp\tAveRMerge\tSpMerR\n");
      printf("--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\tAveRSp\tAveRMerge\tSpMerR\n");
      printf("--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");    
    }
  }
#else
  if(timing_driven ){
    printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\trLim\tTExp\n");
    printf("--------------------------------------------------------------------------------------------------\n");
  }
  else{
    printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tAveAccDCost\tAveNZAccDCost\tAveRejDCost");
    printf("--------------------------------------------------------------------------------------------------\n");    
  }
#endif

  while(!done){
    movesRejected = 0;
    nonZeroMovesAccepted = 0;

    averageDeltaCostAccept = 0.0;
    averageDeltaCostNonZeroAccept = 0.0;
    averageDeltaCostReject = 0.0;

    deltaCostAccept = 0.0;
    deltaCostReject = 0.0;
    
    numMoveLogic = 0; numRegSwap = 0; numRegSplit = 0; numRegMerge = 0; 
    numMoveLogicReject = 0; numRegSwapReject = 0; numRegSplitReject = 0; numRegMergeReject = 0; 
    numZip = 0; numUnzip = 0; numZipReject = 0; numUnzipReject = 0;
    aveRegSplit = 0.0; aveRegMerge = 0.0;
    numMoveRegRetime = 0;
    aveRegRetime = 0.0;
    numRegRetimeReject = 0;
    aveCostRegRetime = 0.0;

    numMovesFailed = 0;

    averageOverallCost = 0.0;
    stdDevOverallCost = 0.0;

    numAttemptedLUTMoves = 0;
    numAttemptedRegMoves = 0;
    numAttemptedIOMoves = 0;
    //averageAttemptedLUTMoveManhattanDistance = 0;
    //averageAttemptedRegMoveManhattanDistance = 0;
    //averageAttemptedIOMoveManhattanDistance = 0;

    if(timing_driven){
      netlist->overallCost = 1.0;
    }
    else{
      netlist->overallCost = netlist->wireCost;
    }

    for(int i = 0; i < numMovesPerTemp; i++){
#if defined(PARANOID_CHECK_PIPEPLACE)
      reallyOldWireCost = netlist->wireCost;
      reallyOldTimingCost = netlist->timingCost;
      reallyOldMaxCut = netlist->maxCut;
      reallyOldAveCut = netlist->aveCut;
#endif

#if defined(PIPEPLACEDEBUGHEAVY)
      cout << "Wire Cost before move = " <<  netlist->wireCost <<endl;
      cout << "Timing Cost before move = " <<  netlist->timingCost <<endl;
      if(maxCutCost){
	cout << "Max Cut before move = " <<  netlist->maxCut <<endl;
      }
      if(aveCutCost){
	cout << "Ave Cut before move = " <<  netlist->aveCut <<endl;
      }
      //netlist->print_netList(true);
      numberMoves++;
      
      if(numberMoves == 4185){
 	cout << endl << "Before move!" << endl;
	netlist->printPlacement();
	netlist->print_netList(true);
      }
#endif

      //Generate a random move
      moveType = generatePipelineAwareMove(netlist, &block0, &block1, &indexOrBank0, &indexOrBank1, &blockType, &numMove, (int) (rLim + 0.5));
      
      //Make the move and update all the proper sink delays.
      //We are also making incremental changes to the slacks - they are inexact, but we'll see what
      // we can do with them
      //We should keep track of if the move failed or not.
      //If it failed, we should not try to undo it.
      moveSucceeded = makePipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, 
					 blockType, numMove, criticalPathDelay, &retimeUndoList);

#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

      //Figure out the cost of the move
      if(deltaDelayCost){
	dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
					     timing_driven, criticalPathDelay, &retimeUndoList);
      }
      else{
	dCost = determineDeltaCost(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				   timing_driven, criticalPathDelay, &retimeUndoList);
      }

#if defined(PIPEPLACEDEBUGHEAVY)
      if(numberMoves == 4185){
 	cout << endl<< "After move!" << endl;
 	netlist->printPlacement();
   	netlist->print_netList(true);
      }
#endif

#if defined(PIPEPLACEDEBUG)      
      if(moveType == MOVELOGIC){
	numMoveLogic ++;
      }
      else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP){
	numRegSwap ++;
      }
      else if(moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST){
	numRegSplit ++;
	aveRegSplit += numMove;
      }
      else if(moveType == MOVEREGPARENT || moveType == MOVEREGCHILD){
	//This isn't exactly merging since we might or might not be trying to push all of the registers back together
	numRegMerge ++;
	aveRegMerge += numMove;
      }
      else if(moveType == MOVEZIPPNET){
	numZip ++;
      }
      else if (moveType == MOVEUNZIPPNET){
	numUnzip ++;
      }
      else if (moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARD){
	numMoveRegRetime ++;
	aveRegRetime += numMove;
	aveCostRegRetime += dCost;
      }
      //else if (moveType == MOVEREGRETIMEBACKABORT || moveType == MOVEREGRETIMEFORWARDABORT){
      //} 
      if(!moveSucceeded){
	numMovesFailed++;
      }
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
      oldWireCost = netlist->wireCost;
      oldTimingCost = netlist->timingCost;
      oldMaxCut = netlist->maxCut;
      oldAveCut = netlist->aveCut;
      
      //Recalculate everthing from scratch
      netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
      if(fabs(oldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
	ERROR("Problem doing incremental cost update");
      }
      if(fabs(oldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
	ERROR("Problem doing incremental cost update");
      }
      if(oldMaxCut != netlist->maxCut){
	ERROR("Problem doing incremental cost update");
      }
      if(oldAveCut != netlist->aveCut){
	ERROR("Problem doing incremental cost update");
      }
#endif
      
      //Figure out whether we want to reject this move
      //This is actually redundant since the dCost will be 0 if the move failed.
      if(dCost > 0.0 && moveSucceeded){
	if(drand48() >= exp(-1.0*(dCost/currTemp))){
	  deltaCostReject += dCost;
	  movesRejected++;

#ifdef PIPEPLACEDEBUG
	  if(moveType == MOVELOGIC){
	    numMoveLogicReject ++;
	  }
	  else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP){
	    numRegSwapReject ++;
	  }
	  else if(moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST){
	    numRegSplitReject ++;
	  }
	  else if(moveType == MOVEREGPARENT || moveType == MOVEREGCHILD){
	    //This isn't exactly merging since we might or might not be trying to push all of the registers back together
	    numRegMergeReject ++;
	  }
	  
	  else if(moveType == MOVEZIPPNET){
	    numZipReject ++;
	  }
	  else if (moveType == MOVEUNZIPPNET){
	    numUnzipReject ++;
	  }	  
	  else if (moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARD){
	    numRegRetimeReject++;
	  }
#endif

	  //Undo the move
	  //cout << "Undoing move of type" << moveType << endl;
	  undoPipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				criticalPathDelay, &retimeUndoList);
    
	  //Calculate the original cost back again
	  if(deltaDelayCost){
	    //dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, 
	    //				 numMove, timing_driven, criticalPathDelay, &retimeUndoList);
	    dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block1, &block0, indexOrBank1, indexOrBank0, blockType, 
						 numMove, timing_driven, criticalPathDelay, &retimeUndoList);
	  }
	  else{
	    //dCost = determineDeltaCost(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
	    //		       timing_driven, criticalPathDelay, &retimeUndoList);
	    dCost = determineDeltaCost(netlist, moveType, &block1, &block0, indexOrBank1, indexOrBank0, blockType, numMove, 
				       timing_driven, criticalPathDelay, &retimeUndoList);
	  }
	  
#ifdef PLACER_DEBUG
	  numRegistersEstimatedFreed = 0;
	  numRegistersEstimatedNeeded = 0;
#endif

#ifdef PIPEPLACEDEBUGHEAVY
 	  if(numberMoves == 4185){
 	    cout << endl << "After undo move!" << endl;
 	    netlist->printPlacement();
 	    netlist->print_netList(true);
 	  }
#endif

	  //If we unzipped a net, then undid it by zipping a net, we've still got info that we don't need anymore
	  // sitting in block 1.
	  // We should go ahead and delete those registers and nets now.
	  if(moveType == MOVEZIPPNET){
	    ((registerBank*)block1)->freeBank();
	  }

	  //If we just retimed registers backwards, then we wanted to undo this move, we created
	  // a new series of "undo" nets and registers.  We should delete the nets and registers
	  // we don't need anymore looking forward and confirm the registers backwards.
	  //This is essentially because we "retimed" register forwards.
	  else if(moveType == MOVEREGUNDOREGRETIMEBACK){
	    clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	    clearNerfedBlocksForward(block0, blockType);
	  }
	  
	  //If we just retimed registers forwards, we should delete the nets and registers
	  // we don't need anymore looking backward and confirm the registers forwards
	  else if(moveType == MOVEREGUNDOREGRETIMEFORWARD){
	    clearRetimeUndoListForward(&retimeUndoList);
	    clearNerfedBlocksBackward(block0, blockType);
	  }

#if defined(PARANOID_CHECK_PIPEPLACE)
	  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
	  if(fabs(reallyOldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
	    ERROR("Problem doing undo cost update");
	  }
	  if(fabs(reallyOldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
	    ERROR("Problem doing undo cost update");
	  }
	  if(reallyOldMaxCut != netlist->maxCut){
	    ERROR("Problem doing incremental cost update");
	  }
	  if(reallyOldAveCut != netlist->aveCut){
	    ERROR("Problem doing incremental cost update");
	  }
#endif
	
	}
	else{
	  netlist->overallCost += dCost;
	  averageOverallCost += netlist->overallCost;

	  deltaCostAccept += dCost;
	  //cout << "Accepted move with dCost = " << dCost <<endl;
	  nonZeroMovesAccepted ++;

	  //If we just zipped a net, we should go ahead and delete the registers in block1 now
	  // we kept them around because of incremental cost calculation and just in case
	  // we had to undo the move, but now we can get rid of them
	  if(moveType == MOVEZIPPNET){
	    ((registerBank*)block1)->freeBank();
	  }

	  //If we just retimed registers backwards, we should delete the nets and registers
	  // we don't need anymore looking forward and confirm the registers backwards
	  else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
	    clearRetimeUndoListForward(&retimeUndoList);
	    clearNerfedBlocksBackward(block0, blockType);
	  }
	  
	  //If we just retimed registers forwards, we should delete the nets and registers
	  // we don't need anymore looking backward and confirm the registers forwards
	  else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
	    clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	    clearNerfedBlocksForward(block0, blockType);
	  }
	}
      }
      else{
	if(dCost < 0.0){
	  nonZeroMovesAccepted ++;
	  //cout << "Accepted move with dCost = " << dCost <<endl;
	}

	netlist->overallCost += dCost;
	averageOverallCost += netlist->overallCost;

	deltaCostAccept += dCost;

	//If we just zipped a net, we should go ahead and delete the registers in block1 now
	// we kept them around because of incremental cost calculation and just in case
	// we had to undo the move, but now we can get rid of them
	if(moveType == MOVEZIPPNET){
	  ((registerBank*)block1)->freeBank();
	}

	//If we just retimed registers backwards, we should delete the nets and registers
	// we don't need anymore looking forward and confirm the registers backwards
	else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
	  clearRetimeUndoListForward(&retimeUndoList);
	  clearNerfedBlocksBackward(block0, blockType);
	}
	
	//If we just retimed registers forwards, we should delete the nets and registers
	// we don't need anymore looking backward and confirm the registers forwards
	else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
	  clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	  clearNerfedBlocksForward(block0, blockType);
	}
      }

      //To be completely accurate, we should now recalculate the critical path delay and slacks
      // for the whole system.
      //However, this is too timing consuming to do after each move, so we don't do it all of the time
      // we only do it after every N moves
      if(i % numMovesBeforeUpdate == (numMovesBeforeUpdate - 1)){
	criticalPathDelay =  calculateCriticalPathDelay(netlist, true);
	//cout << "Critical Path updated to " << criticalPathDelay << endl;

	//Now, the costs for all of the nets might have changed, so calculate the placement cost from scratch
	netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
      }

#if defined(PIPEPLACEDEBUGHEAVY)
      cout << "Wire Cost after move = " <<  netlist->wireCost <<endl;
      cout << "Timing Cost after move = " <<  netlist->timingCost <<endl;
      if(maxCutCost){
	cout << "Max Cut after move = " <<  netlist->maxCut <<endl;
      }
      if(aveCutCost){
	cout << "Ave Cut after move = " <<  netlist->aveCut <<endl;
      }
      //netlist->print_netList();
      //netlist->printPlacement();
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
      //FPGA->checkFreeRegisters();
      netlist->checkNets();
      if(numRegistersEstimatedFreed != 0){
	//ERROR("Did not estimate # of freed registers correctly");
      }
      if(numRegistersEstimatedNeeded != 0){
	//ERROR("Did not estimate # of needed registers correctly");
      }
#endif
    }

    //Calculate some statistics about this iteration
    //netlist->printPlacement();
    
    if(movesRejected > 0){
      averageDeltaCostReject = deltaCostReject / (double)(movesRejected);
    }
    if(numMovesPerTemp - movesRejected > 0){
      averageDeltaCostAccept = deltaCostAccept / (double)(numMovesPerTemp - movesRejected);
    }
    if(nonZeroMovesAccepted > 0){
      //cout << "Total delta Accept = " << deltaCostAccept << endl;
      averageDeltaCostNonZeroAccept = deltaCostAccept / (double) nonZeroMovesAccepted;
      //cout << "Accepted " << nonZeroMovesAccepted << " nonzero moves with an average of " << averageDeltaCostNonZeroAccept << endl;
    }
    if(numMoveRegRetime > 0){
      aveRegRetime = aveRegRetime / numMoveRegRetime;  
      aveCostRegRetime = aveCostRegRetime / numMoveRegRetime;
    }

    averageOverallCost /= numMovesPerTemp - movesRejected;

#if defined (PIPEPLACEDEBUG)
    aveRegSplit = aveRegSplit / numRegSplit;
    aveRegMerge = aveRegMerge / numRegMerge;
    if(timing_driven){
      if((numRegSplit - numRegSplitReject) < 0){
	ERROR("SOMETHING HAPPENED HERE!");
      }
      if(maxCutCost && !aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      
      else if(aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%f\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else{
// 	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%d\t%d\t%d\t%d", 
// 	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
// 	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
// 	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, (float) numMoveLogic / (float) numMovesPerTemp,
// 	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
// 	       (float) numZip / (float) numMovesPerTemp,(float) numUnzip / (float) numMovesPerTemp,
// 	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
// 	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
// 	       (float)(numZip - numZipReject) / (float) numZip, (float)(numUnzip - numUnzipReject) / (float) numUnzip,
// 	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject),
// 	       numRegSplit, numRegSplitReject, numRegMerge, numRegMergeReject);

	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%1.2f\t%1.2f\t%d\t%d\t%1.2f\t%1.4e\t%d", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, numMoveLogic,
	       numRegSwap, numRegSplit, numRegMerge, numZip, numUnzip,(numMoveLogic - numMoveLogicReject), (numRegSwap - numRegSwapReject),
	       (numRegSplit - numRegSplitReject), (numRegMerge - numRegMergeReject),
	       (numZip - numZipReject), (numUnzip - numUnzipReject),
	       aveRegSplit, aveRegMerge,  numMoveRegRetime, numRegRetimeReject, aveRegRetime, aveCostRegRetime, numMovesFailed);
      }
    }
    else{
      if(maxCutCost && !aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else if(aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%f\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else{
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
   }
#else
    if(timing_driven){
      if(maxCutCost && !aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%e\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost, 
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut);
      }
      
      else if(aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut);
      }
      else{
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, rLim, tdPlacementCritExponent);
      }
    }
    else{
      if(maxCutCost && !aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut);
      }
      else if(aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut);
      }
      else{
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject);
      }
    }
#endif

    cout << endl;

    //Calculate the average, max and min criticalities for all pin-to-pin connections
    netlist->calculateAvgMaxMinCrits(criticalPathDelay);
    

    //printf("Num Splits = %d, Num Split Attempts = %d, NumSplitSwap = %d, Num Merges = %d, Num Merge Attempts = %d, numMergeSwaps = %d \n", 
    //numSplits, numSplitCalls, numSplitSwapCalls, numMerges, numMergeCalls, numMergeSwapCalls);

    //averageAttemptedIOMoveManhattanDistance /= numAttemptedIOMoves;
    //averageAttemptedLUTMoveManhattanDistance /= numAttemptedLUTMoves;
    //averageAttemptedRegMoveManhattanDistance /= numAttemptedRegMoves;
    //cout << "Average Attempted IO move Manhattan distance = " <<  averageAttemptedIOMoveManhattanDistance << " over " << numAttemptedIOMoves << " number of IO moves" << endl;
    //cout << "Average Attempted LUT move Manhattan distance = " <<  averageAttemptedLUTMoveManhattanDistance << " over " << numAttemptedLUTMoves << " number of LUT moves" << endl;
    //cout << "Average Attempted Reg move Manhattan distance = " <<  averageAttemptedRegMoveManhattanDistance << " over " << numAttemptedRegMoves << " number of Reg moves" << endl;

    //Update the temperature, range limit and timing criticality exponent
    currTemp = updateTemp(currTemp, numMovesPerTemp - movesRejected, numMovesPerTemp);
    rLim = updateRLimAndCritExp(rLim, numMovesPerTemp - movesRejected, numMovesPerTemp);

    //Re-calculate the critical path delay
    criticalPathDelay =  calculateCriticalPathDelay(netlist, true);

    //Update the cost of the current placement due to new critical path delay and new 
    netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

    //See if we should terminate the annealing
    done = exitCondition(currTemp, netlist->overallCost, netlist->calculateNumNetsAndPNets());
    //done = true;
  }
  
  cout << "Final Wire Cost " << netlist->wireCost << endl;
  cout << "Final placement Timing Cost" << netlist->timingCost << endl;
  cout << "Final Critical Path Delay " << criticalPathDelay << endl;
  //netlist->printNetWireCost(true);
}

double updateTemp(double currTemp, int movesAccepted, int movesPerTemp){
  double accPercent = (double)movesAccepted / (double) movesPerTemp;
  if(accPercent > 0.96){
    return (currTemp * 0.5);
  }
  else if(accPercent > 0.8){
    return (currTemp * 0.9);
  }
  if(accPercent > 0.15){
    return (currTemp * 0.95);
  }
  else{
    return (currTemp * 0.8);
  }
}

//Figure out if we should terminate the annealing
bool exitCondition(double currTemp, double cost, int numNets){
  //We are done if the temperature is less than 0.005 * cost of an average net

  if(currTemp < 0.005 * cost / (double) numNets){
    return true;
  }
  return false;    
}

//This updates the Rlim and timing criticality exponent
float updateRLimAndCritExp(float rLim, int movesAccepted, int movesPerTemp){
  float newRLim;
  float accPercent = (float) movesAccepted / (float) movesPerTemp;
  float maxMove;
  
  //This is the range limit mode suggested by VPR
  //As per VPR -
  //<quote>
  /* Update the range limited to keep acceptance prob. near 0.44.  Use *
   * a floating point rlim to allow gradual transitions at low temps.  */
  //</quote>

  maxMove = (float) max(FPGA->get_numRows(), FPGA->get_numCols());  

  if(noWindowing){
    newRLim = maxMove;
  }
  else{
    //Don't let the range get larger than the device
    newRLim = flomin(rLim * (0.56 + accPercent), maxMove);
    
    //Don't let the range get smaller than 1
    newRLim = flomax(newRLim, 1.0);
  }

  tdPlacementCritExponent = (1.0 - ((newRLim - 1.0) / (maxMove - 1.0))) * (tdPlacementCritExponentFinal - 1.0) + 1.0;

  return newRLim;
}

//Go through the netlist and find registers that are not in the same CLB as their host LUT.
//If they are not in the same CLB, try and move it there.
//1) If the current CLB has a free register, just put it there.
//2) If it doesn't, swap positions with an unrelated FF in that CLB (there must be at least one)
void findOrphanRegisters(Netlist *netlist, bool timing_driven){
  void* block0;
  void* block1;
  Terminal *outTerm;
  Terminal *sink;
  logicBlock *sourceLB;
  int sourceRow, sourceCol;
  int registerRow, registerCol;
  double criticalPathDelay;
  double dCost;
  int moveType;
  int numRegisters = 0;
  int numPackableRegisters = 0;
  int numOrphanedRegisters = 0;
  int minSinkDistance;

  double oldWireCost;
  double oldTimingCost;
  double oldCriticalPathDelay;

  if(timing_driven){
    netlist->overallCost = 1.0;
  }
  else{
    netlist->overallCost = netlist->wireCost;
  }

  criticalPathDelay = calculateCriticalPathDelay(netlist, true);

  for(list<logicBlock*>::iterator lbIter = netlist->lbList.begin(); lbIter != netlist->lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() == REGISTER_BANK){
      numRegisters++;
      registerRow = (*lbIter)->row;
      registerCol = (*lbIter)->col;
      
      //Check to see if it's source is a LUT
      if((*lbIter)->returnDriverType() == LUT){
	cout << "Register " << (*lbIter)->name;
	numPackableRegisters++;

	sourceLB = (*lbIter)->getRegisterSource();

	sourceRow = sourceLB->row;
	sourceCol = sourceLB->col;

	//Show how far this register is away from it's sinks
	//(*lbIter)->printDistanceToAllSinks();
	  
	if(sourceCol != registerCol || sourceRow != registerRow){
	  //This register is not in the same CLB as it's parent LUT
	  numOrphanedRegisters++;
	  //cout << " is orphaned at (" << registerCol << ", " << registerRow;
	  //cout << "), instead of with source LUT at (" << sourceCol << ", " << sourceRow << ")" << endl;	  
	  //cout << "Distance to parent LUT = " << abs(sourceCol - registerCol) + abs(sourceRow - registerRow) <<endl;
	  
	  cout << " orphaned with minSinkDistance = " << (*lbIter)->minDistanceToAllSinks();
	  cout << " and distance to parent LUT = " << abs(sourceCol - registerCol) + abs(sourceRow - registerRow) <<endl;

	  block0 = (void*)(*lbIter)->rb;

	  //Find a register in the host LUT's CLB to swap with
	  block1 = (void *) FPGA->getCLBAt(sourceRow, sourceCol)->getFirstEmptyUnrelatedRegisterBank();
	  
	  if(block1){
	    oldWireCost = netlist->wireCost;
	    oldTimingCost = netlist->timingCost;
	    oldCriticalPathDelay = criticalPathDelay;

	    moveType = MOVELOGIC;
	    makePipelineAwareMove(&moveType, netlist, &block0, &block1, 0, 0, 
				  REGISTERBANK, 0, criticalPathDelay, NULL);
	    
#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

	    //Figure out the cost of the move
	    if(deltaDelayCost){
	      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, 0, 0, REGISTERBANK, 0, 
						   timing_driven, criticalPathDelay, NULL);
	    }
	    else{
	      dCost = determineDeltaCost(netlist, moveType, &block0, &block1, 0, 0, REGISTERBANK, 0, 
					 timing_driven, criticalPathDelay, NULL);
	    }
	    netlist->overallCost += dCost;

	    criticalPathDelay = calculateCriticalPathDelay(netlist, true);

	    cout << "Move to parent delta cost = " << dCost << endl;
	    cout << "Move to parent delta wire cost = " << netlist->wireCost - oldWireCost << endl;
	    cout << "Move to parent delta timing cost = " << netlist->timingCost - oldTimingCost << endl;
	    cout << "Move to parent delta critical path delay = " << criticalPathDelay - oldCriticalPathDelay << endl;

	    //Undo the move
	    undoPipelineAwareMove(&moveType, netlist, &block0, &block1, 0, 0, REGISTERBANK, 0, 
				  criticalPathDelay, NULL);
	    
	    //Calculate the original cost back again
	    if(deltaDelayCost){
	      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block1, &block0, 0, 0, REGISTERBANK, 0, 
						   timing_driven, criticalPathDelay, NULL);
	    }
	    else{
	      dCost = determineDeltaCost(netlist, moveType, &block1, &block0, 0, 0, REGISTERBANK, 0, 
					 timing_driven, criticalPathDelay, NULL);
	    }
	    netlist->overallCost += dCost;

	    criticalPathDelay = calculateCriticalPathDelay(netlist, true);
	  }
	}
	else{
	  cout << " united with parent minSinkDistance = " << (*lbIter)->minDistanceToAllSinks() << endl;
	}
      }
    }
  }
  cout << numOrphanedRegisters << " registers out of a total of " << numRegisters << " registers and ";
  cout << numPackableRegisters << " packable registers were orphaned in the regular placement" << endl;
}
