// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
//extern long int wireCost;
extern long int sharingCost;
extern vprFPGA *FPGA;
extern int MAXLATENCY;
extern int MAXID;

#ifdef ROUTER_DEBUG
extern long routerPops;
extern long routerPushes;
#endif

//This is the primary entry point for the congestion-driven pipelined router.
void routeNetQuickRoute(int iter, Net* net){
  PipeliningNet *pNet;
  //char tempStr [MAX_STR_LEN];
  string tempStr;

  list<PipeliningNet *>::iterator pNetIter;
  priority_queue<pqNode>* PQ;
  list<pExplore> cleanupList;
  vector <unsigned int> *rootPathVector;
  bool success;
  int sinkCount;
  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;

  int failedLatency;

  //Check to make sure that all of the routing nodes are un-explored/etc before we start
#ifdef PARANOID_CHECK_ROUTE_PIPELINE
  checkEmptyRRGraph(true);
#endif
  
  //First, we should create a pipelined net for the source and put it in the list
  //strcpy(tempStr, net->name);
  //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
  tempStr = net->name + "PipelineNet", itoa(net->pipelineNetList.size());
  pNet = new PipeliningNet(tempStr, net->source, 0, net);
  net->pipelineNetList.push_back(pNet);

  //Make the source terminal point to this new pipelined net
  pNet->source->destPipeliningNet = pNet;

  //Put the source node into the routing tree of the first pipelined net
  //and increment the occupancy
  pNet->init_rTreeRoot();

  //Increment occupancy and put it into the tree at latency 0
  if(pNet->source->rNode->get_type() != DELAY){
    pNet->source->rNode->set_occupancy(pNet->source->rNode->occupancy + 1);
    if(pNet->source->rNode->get_occupancy() > pNet->source->rNode->get_capacity()){
      sharingCost+=1;
    }
  }

  pNet->source->rNode->set_inPRTree(0);

  //Make a new PQ
  PQ=new priority_queue<pqNode>();
      
  //Congestion-driven routing finds sinks in no particular order
  //Loop until we fail routing or until we find all of our sinks
  sinkCount = 0;
  while(1){
    //net->print_Info();

    //Check to make sure that all of the routing nodes are un-explored (although not the inPRTree)
#ifdef PARANOID_CHECK_ROUTE_PIPELINE
    checkEmptyRRGraph(false);
#endif

    //Make sure that the PQ is empty
    if(!PQ->empty()){
      ERROR("PQ not empty!");
    }

    //Add all of the existing routing trees into the PQ
    //cout << "Adding existing paths to PQ" << endl;
    for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
      //cout << "\tCreating base pathvector for pipelining net " << (*pNetIter)->name << endl;
      rootPathVector = createPathVectorRoot(net, *pNetIter, &cleanupList);
      //cout << "\tAdding nodes of pipelining net " << (*pNetIter)->name << " to PQ" << endl;
      addRoutingNodes2PQ(PQ, (*pNetIter), (*pNetIter)->rTree.root, (*pNetIter)->latency, rootPathVector, &cleanupList);
      delete rootPathVector;
    }
      
    //Find a sink at the appropriate latency and backtrace/add sinks to appropriate pipelined nets
    //cout << "Begin Routing" << endl;
    success = findRouteQuickRoute(iter, net, PQ, &cleanupList);
      
    if(success){
      sinkCount++;
    }
    else{
      //OK, the router got stuck, so figure out what sinks have not been routed yet
      cout << "Primary router failed, invoking secondary router" << endl;
	
      for(sinkIter = net->sinkVector.begin(); sinkIter != net->sinkVector.end(); sinkIter++){
	if(!(*sinkIter)->isRouted){	
	  cout << "Router has not routed sink " << (*sinkIter)->name << endl;
	}
      }
      for(sinkIter = net->sinkVector.begin(), latencyIter = net->latencyVector.begin(); 
	  sinkIter != net->sinkVector.end(), latencyIter != net->latencyVector.end(); 
	  sinkIter++, latencyIter++){
	if(!(*sinkIter)->isRouted){
	  //Try to route to this sink from decreasing latency existing routes
	  //At the very worse, we can always try routing again from the 0 latency route
	  //This will almost always allow for a N latency route if one exists...
	  cout << "Attempting to route to " << (*sinkIter)->name << " at latency " << *latencyIter << endl;

	  failedLatency = *latencyIter;
	  while(!success){
	    //Make sure that the PQ is empty
	    if(!PQ->empty()){
	      ERROR("PQ not empty!");
	    }
	      
	    cout << "Attempting to begin from " << failedLatency << " latency nets" << endl;

	    //Add only the N latency nets to the PQ
	    for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
	      if((*pNetIter)->latency == failedLatency){ 
		cout << "\tAdding nodes of pipelining net " << (*pNetIter)->name << " to PQ" << endl;
		rootPathVector = createPathVectorRoot(net, *pNetIter, &cleanupList);		
		addRoutingNodes2PQ(PQ, (*pNetIter), (*pNetIter)->rTree.root, (*pNetIter)->latency, rootPathVector, &cleanupList);
		delete rootPathVector;
	      }
	    }

	    //Try to find this particular sink
	    success = findRouteQuickRouteSink(iter, net, PQ, &cleanupList, (*sinkIter), (*latencyIter));
	      
	    if(success){
	      cout << "Succeeded!" << endl;
	      sinkCount++;
	      break;
	    }
	    else{
	      cout << "Failed!" << endl;
	      if(failedLatency > 0){
		failedLatency--;
	      }
	      else{
		cout << "Failed from all latencies!  I'm out of suggestions!" << endl;
		ERROR("Failed routing!");
	      }
	    }
	  }
	}
      }
    }
      
    if(sinkCount == net->sinkVector.size()){
      //We are done routing so we should reset all of the sinks to unrouted
      for(sinkIter = net->sinkVector.begin(); sinkIter != net->sinkVector.end(); sinkIter++){
	(*sinkIter)->isRouted = false;
      }

      //Go through all of the pipelining nets and un-set the inRoutingTree variables
      for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
	(*pNetIter)->rTree.clearInPRTree((*pNetIter)->rTree.root);
      }
	
      //cout << "Net completed routing!" << endl;
      break;
    }
  }
  delete PQ;
}

//Find a congestion-driven route from the given PQ to any undiscovered sink.
bool findRouteQuickRoute(int iter, Net* net, priority_queue<pqNode>* PQ, list<pExplore> *cleanupList){
   PipeliningNet* pNet;

  routingNode* currentRNode;
  int currentLatency;
  int currentExploreIndex;
  vector <unsigned int> *currentPathVector;  

  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;
  routingNode* nb_rnode;
  routingEdge* nb_edge;
  vector <unsigned int> *nb_pathVector;  

  double path_cost;
  double delay;
  double new_path_cost;
  double new_delay;
  
  pExplore *prev;

  //Try to route to the sink
  while(1){
    //Grab something usable off of the PQ
    while(1){
      if(PQ->empty()){
	//We have emptied the PQ before we found the sink
	cout << "Could not find route for net " << net->name << endl;

	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(0).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}

	return false;
      }
      
      currentRNode=(PQ->top()).rNode;
      currentLatency = PQ->top().latency;
      currentExploreIndex = currentRNode->pIsExplored.at(0).at(currentLatency);
      
      //Determine if we have explored the current node K times already at the given latency
      if(currentExploreIndex >= KFUNCTION(currentLatency)){
#if defined(ROUTER_DEBUG)
	cout << "\tPopped " << currentRNode->get_name() <<" predecessor ";
	if((PQ->top()).pred.rNode){
	  cout<<(PQ->top()).pred.rNode->get_name();
	  cout << " AC index 0 " << " latency " << (PQ->top()).pred.latency;
	  cout << " and e_Index " << (PQ->top()).pred.exploreIndex; 
	}
	cout << " but node has been explored at latency " << currentLatency << " " << currentExploreIndex << " times" << endl;
#endif
	delete PQ->top().pathVector;
	PQ->pop();
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	continue;
      }
      else{
	//This node is free to be explored
	currentPathVector = PQ->top().pathVector;  

	//Mark node explored once more at current latency and set pathcost/delay/predecessor
	currentRNode->pIsExplored.at(0).at(currentLatency)++;
	currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pathCost;
	currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).netDelay;
	currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pred;
	addNode2PathVector(currentRNode, currentPathVector, cleanupList);
	
	cleanupList->push_back(pExplore(currentRNode, 0, currentLatency, currentExploreIndex));
		
#if defined(ROUTER_DEBUG)
	cout <<"\tPopped "<<currentRNode->get_name();
	cout << " at path cost "<< currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
	cout << " has been explored at latency " << currentLatency << " " << currentExploreIndex << " times has delay ";
	cout << currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) << " and predecessor node ";
	if(currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).rNode){
	  cout << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).rNode->get_name();
	  cout << " AC index 0";
	  cout << " latency " << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).latency;
	  cout << " and e_Index " << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).exploreIndex;
	}
	cout << endl;
#endif

	pNet = PQ->top().sourcePNet;

	PQ->pop();	
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	break;
      }
    }

    //We've got something usable off of the PQ, so now see if we have hit a new sink at the correct latency
    for(sinkIter = net->sinkVector.begin(), latencyIter = net->latencyVector.begin(); 
	sinkIter != net->sinkVector.end(), latencyIter != net->latencyVector.end(); 
	sinkIter++, latencyIter++){
      if(!(*sinkIter)->isRouted && currentRNode == (*sinkIter)->rNode && currentLatency == (*latencyIter)){
#if defined(ROUTER_DEBUG)
	cout << "\tFOUNDSINK " <<  currentRNode->name << " at latency " << currentLatency;
	cout << " at cost " << currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
	cout << " and delay " << currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) << endl <<endl;
#endif

	//Mark sink as routed
	(*sinkIter)->isRouted = true;

	//Backtrace and create the appropriate pipelining nets, etc.
	backTracePipelinedRoute(net, pNet, (*sinkIter), currentLatency, currentExploreIndex);

	delete currentPathVector;
	
	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(0).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}
	MAXID = 0;

	//We need to free the rest of the path vectors that we pushed onto the queue
	while(!PQ->empty()){
	  delete PQ->top().pathVector;
	  PQ->pop();
	}

	//Exit
	return(true);
      }
    }

    //We didn't hit a sink at the correct latency, so we should add neighbors to PQ
    currentRNode->create_fanoutListItr();
    for(nb_edge=currentRNode->getnext_fanoutList(); nb_edge!=0; nb_edge=currentRNode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      
#if defined(ROUTER_DEBUG)
      cout<<"\tLooking at neighbor node "<<nb_rnode->get_name() << endl;
#endif
      
      //Only add explore-able neighbors to the queue
      //Don't add the neighbor if-
      //1) Neighbor is in explored path to current node
      if(checkNodeID(currentPathVector, nb_rnode->nodeID)){
	//cout << "\t\tNeigbor not added since already in current path" << endl;
        continue;
      }

      //2) Neighbor has already been explored K times at latency or once already at latency 0
      if(nb_rnode->pIsExplored.at(0).at(currentLatency) >= KFUNCTION(currentLatency)){
	//cout << "\t\tNeighbor not added since already explored " << nb_rnode->pIsExplored.at(0).at(currentLatency) << " times" << endl;
	continue;
      }

      //3) Neighbor is a delay node and we don't have any sinks in this net that require a latency that high
      if(nb_rnode->type == DELAY && currentLatency == net->netMaxLatency){
	//cout << "\t\tNeighbor not added since we do not need to search above latency " << currentLatency << endl;
	continue;
      }
      
      //4) Neighbor is already in tree at given latency
      if((nb_rnode->type != DELAY && nb_rnode->check_inPRTree(currentLatency)) ||
	 (nb_rnode->type == DELAY && nb_rnode->check_inPRTree(currentLatency + 1))){

	if(nb_rnode->type == DELAY){
	  //cout << "\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency + 1 << endl;
	}
	else{
	  //cout << "\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency << endl;
	}
	continue;
      }

      //Add the neighbor to the PQ
      path_cost = currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
      
      if(currentRNode->type == DELAY){
	delay = currentRNode->intrinsicDelay;;
      }
      else{
	delay = currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex);
      }
      
      new_path_cost=path_cost+nb_rnode->returnCongestionCost(iter);

      //Make a new path vector for the neighbor
      nb_pathVector = new vector <unsigned int> (*currentPathVector);  

      if(nb_rnode->type == DELAY){
	//Delay nodes get explored at latency L + 1 and do not add it their own intrinsic delay
	new_delay = delay + nb_edge->intrinsicDelay;
	PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, 0, currentLatency + 1, 
			pExplore(currentRNode, 0, currentLatency, currentExploreIndex), nb_pathVector));
#ifdef ROUTER_DEBUG
	routerPushes++;
	cout <<"\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	cout << " latency " << currentLatency + 1;
	cout << " and predecessor "<< currentRNode->get_name();
	cout << " AC index 0 ";
	cout << " latency " << currentLatency;
	cout << " and e_Index " << currentExploreIndex << endl;
#endif	  	  
      }
      else{
	new_delay = delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, 0, currentLatency, 
			pExplore(currentRNode, 0, currentLatency, currentExploreIndex), nb_pathVector));
#ifdef ROUTER_DEBUG
	routerPushes++;
	cout <<"\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	cout << " latency " << currentLatency;
	cout << " and predecessor "<< currentRNode->get_name();
	cout << " AC index 0 ";
	cout << " latency " << currentLatency;
	cout << " and e_Index " << currentExploreIndex << endl;
#endif	  	  
      }
    }

    currentRNode->remove_fanoutListItr();;

    delete currentPathVector;
  }
}

//Find a route from the given PQ to one particular sink.
bool findRouteQuickRouteSink(int iter, Net* net, priority_queue<pqNode>* PQ, list<pExplore> *cleanupList, Terminal *sinkTerminal, int latency){
   PipeliningNet* pNet;

  routingNode* currentRNode;
  int currentLatency;
  int currentExploreIndex;
  vector <unsigned int> *currentPathVector;  

  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;
  routingNode* nb_rnode;
  routingEdge* nb_edge;
  vector <unsigned int> *nb_pathVector;  

  double path_cost;
  double delay;
  double new_path_cost;
  double new_delay;
  
  pExplore *prev;

  //Try to route to the sink
  while(1){
    //Grab something usable off of the PQ
    while(1){
      if(PQ->empty()){
	//We have emptied the PQ before we found the sink
	cout << "Could not find route for net " << net->name << endl;

	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(0).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}

	return false;
      }
      
      currentRNode=(PQ->top()).rNode;
      currentLatency = PQ->top().latency;
      currentExploreIndex = currentRNode->pIsExplored.at(0).at(currentLatency);
      
      //Determine if we have explored the current node K times already at the given latency
      if(currentExploreIndex >= KFUNCTION(currentLatency)){
#if defined(ROUTER_DEBUG)
	cout << "\tPopped " << currentRNode->get_name() <<" predecessor ";
	if((PQ->top()).pred.rNode){
	  cout<<(PQ->top()).pred.rNode->get_name();
	  cout << " AC index 0 " << " latency " << (PQ->top()).pred.latency;
	  cout << " and e_Index " << (PQ->top()).pred.exploreIndex; 
	}
	cout << " but node has been explored at latency " << currentLatency << " " << currentExploreIndex << " times" << endl;
#endif
	delete PQ->top().pathVector;
	PQ->pop();
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	continue;
      }
      else{
	//This node is free to be explored
	currentPathVector = PQ->top().pathVector;  

	//Mark node explored once more at current latency and set pathcost/delay/predecessor
	currentRNode->pIsExplored.at(0).at(currentLatency)++;
	currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pathCost;
	currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).netDelay;
	currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pred;
	addNode2PathVector(currentRNode, currentPathVector, cleanupList);
	
	cleanupList->push_back(pExplore(currentRNode, 0, currentLatency, currentExploreIndex));
		
#if defined(ROUTER_DEBUG)
	cout <<"\tPopped "<<currentRNode->get_name();
	cout << " at path cost "<< currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
	cout << " has been explored at latency " << currentLatency << " " << currentExploreIndex << " times has delay ";
	cout << currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) << " and predecessor node ";
	if(currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).rNode){
	  cout << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).rNode->get_name();
	  cout << " AC index 0";
	  cout << " latency " << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).latency;
	  cout << " and e_Index " << currentRNode->pPredecessor.at(0).at(currentLatency).at(currentExploreIndex).exploreIndex;
	}
	cout << endl;
#endif

	pNet = PQ->top().sourcePNet;

	PQ->pop();	
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	break;
      }
    }

    //We've got something usable off of the PQ, so now see if we have hit the sink at the correct latency
    if(currentRNode == sinkTerminal->rNode && currentLatency == latency){
#if defined(ROUTER_DEBUG)
	cout << "\tFOUNDSINK " <<  currentRNode->name << " at latency " << currentLatency;
	cout << " at cost " << currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
	cout << " and delay " << currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex) << endl <<endl;
#endif

	//Mark sink as routed
	sinkTerminal->isRouted = true;

	//Backtrace and create the appropriate pipelining nets, etc.
	backTracePipelinedRoute(net, pNet, sinkTerminal, currentLatency, currentExploreIndex);

	delete currentPathVector;
	
	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(0).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(0).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}
	MAXID = 0;

	//We need to free the rest of the path vectors that we pushed onto the queue
	while(!PQ->empty()){
	  delete PQ->top().pathVector;
	  PQ->pop();
	}

	//Exit
	return(true);
    }

    //We didn't hit the sink at the correct latency, so we should add neighbors to PQ
    currentRNode->create_fanoutListItr();
    for(nb_edge=currentRNode->getnext_fanoutList(); nb_edge!=0; nb_edge=currentRNode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      
#if defined(ROUTER_DEBUG)
      cout<<"\tLooking at neighbor node "<<nb_rnode->get_name() << endl;
#endif
      
      //Only add explore-able neighbors to the queue
      //Don't add the neighbor if-
      //1) Neighbor is in explored path to current node
      if(checkNodeID(currentPathVector, nb_rnode->nodeID)){
	//cout << "\t\tNeigbor not added since already in current path" << endl;
        continue;
      }

      //2) Neighbor has already been explored K times at latency or once already at latency 0
      if(nb_rnode->pIsExplored.at(0).at(currentLatency) >= KFUNCTION(currentLatency)){
	//cout << "\t\tNeighbor not added since already explored " << nb_rnode->pIsExplored.at(0).at(currentLatency) << " times" << endl;
	continue;
      }

      //3) Neighbor is a delay node and we don't have any sinks in this net that require a latency that high
      if(nb_rnode->type == DELAY && currentLatency == net->netMaxLatency){
	//cout << "\t\tNeighbor not added since we do not need to search above latency " << currentLatency << endl;
	continue;
      }
      
      //4) Neighbor is already in tree at given latency
      if((nb_rnode->type != DELAY && nb_rnode->check_inPRTree(currentLatency)) ||
	 (nb_rnode->type == DELAY && nb_rnode->check_inPRTree(currentLatency + 1))){

	if(nb_rnode->type == DELAY){
	  //cout << "\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency + 1 << endl;
	}
	else{
	  //cout << "\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency << endl;
	}
	continue;
      }

      //Add the neighbor to the PQ
      path_cost = currentRNode->pPathCost.at(0).at(currentLatency).at(currentExploreIndex);
      
      if(currentRNode->type == DELAY){
	delay = currentRNode->intrinsicDelay;
      }
      else{
	delay = currentRNode->pNetDelay.at(0).at(currentLatency).at(currentExploreIndex);
      }

      new_path_cost=path_cost+nb_rnode->returnCongestionCost(iter); 

      //Make a new path vector for the neighbor
      nb_pathVector = new vector <unsigned int> (*currentPathVector);

      if(nb_rnode->type == DELAY){
	//Delay nodes get explored at latency L + 1 and do not add it their own intrinsic delay
	new_delay = delay + nb_edge->intrinsicDelay;
	PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, 0, currentLatency + 1, 
			pExplore(currentRNode, 0, currentLatency, currentExploreIndex), nb_pathVector));
#ifdef ROUTER_DEBUG
	routerPushes++;
	cout <<"\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	cout << " latency " << currentLatency + 1;
	cout << " and predecessor "<< currentRNode->get_name();
	cout << " AC index 0 ";
	cout << " latency " << currentLatency;
	cout << " and e_Index " << currentExploreIndex << endl;
#endif	  	  
      }
      else{
	new_delay = delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, 0, currentLatency, 
			pExplore(currentRNode, 0, currentLatency, currentExploreIndex), nb_pathVector));
#ifdef ROUTER_DEBUG
	routerPushes++;
	cout <<"\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	cout << " latency " << currentLatency;
	cout << " and predecessor "<< currentRNode->get_name();
	cout << " AC index 0 ";
	cout << " latency " << currentLatency;
	cout << " and e_Index " << currentExploreIndex << endl;
#endif	  	  
      }
    }
    currentRNode->remove_fanoutListItr();;

    delete currentPathVector;
  }
}

//This adds all of the routing nodes in the pipelining net's routing tree to the PQ
//If we are doing congestion-driven routing we'll add it to the PQ at zero path cost.
//This is my version of updatePQ
//One big difference is that this function also needs to fill in the pathvector.
//We are given a root pathvector, but
void addRoutingNodes2PQ(priority_queue<pqNode>* PQ, PipeliningNet* pNet, treeNode* tree_node, 
			int latency, vector <unsigned int> *basePathVector, list<pExplore> *cleanupList){
  treeNode* t=tree_node;
  treeNode* child;
  routingNode* parentRNode;
  vector <unsigned int> *pathVector;
  
  //This net may have nothing in it's routing tree
  if(!t){
    return;
  }

  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");

  if(t->get_parent()){
    parentRNode = t->get_parent()->get_rNode();
  }
  else{
    parentRNode = 0;
  }

  //Add everything in the routing tree to the PQ except for non-source delay nodes
  //Sink delay nodes get added to the PQ by their destination nets
  if(t->get_rNode()->get_type() != DELAY || pNet->source->rNode == t->get_rNode()){
    //cout << "\t\tAdding node " << t->get_rNode()->name << " to PQ" << endl;

    //Make a copy of the base pathvector
    pathVector = new vector <unsigned int> (*basePathVector);
  
    //Add myself into the pathvector
    addNode2PathVector(t->rNode, pathVector, cleanupList);

    //Make sure that I am in the routing tree at the proper latency
    if(!t->rNode->check_inPRTree(latency)){
      ERROR("Node is not in PRouting tree!");
    }

    //If we are routing in congestion mode, push the entire tree at the given latency and a cost of 0
    //As far as the predecessor goes, set it to null
    PQ->push(pqNode(pNet, t->get_rNode(), 0.0, t->linkDelay, 0, latency, pExplore(), pathVector));
#ifdef ROUTER_DEBUG
    routerPushes++;
    cout<<"\t\tPushed "<< t->get_rNode()->get_name() <<" at Path Cost " << 0.0;
    cout<<", delay " << t->linkDelay << " and latency " << latency << endl; 
#endif

    //Add any children to the PQ
    t->create_childListItr();
    for(child=t->getnext_childList();child!=0;child=t->getnext_childList()){
      addRoutingNodes2PQ(PQ, pNet, child, latency, pathVector, cleanupList);
    }
    t->remove_childListItr();
  }
}

//This takes in an L latency route, updates the occupany of the routing nodes and adds the correct terms to the
//applicable nets.
void backTracePipelinedRoute(Net* net, PipeliningNet* sourcePNet, Terminal* sinkTerminal, int lat, int e_index){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0; 
  pExplore* current;
  pExplore* end;

  PipeliningNet *currPNet;
  PipeliningNet *nextPNet;
  Terminal* currSink;
  //char tempStr [MAX_STR_LEN];
  string tempStr;

  Terminal* currTerm;
  Terminal* nextTerm;

  //cout << "Backtrace" << endl;  
  //cout << "\tPath back to source or source net" << endl;
  end = new pExplore(sinkTerminal->rNode, 0, lat, e_index);
  current = end;
  //btpr(current);

  //Figure out what pipelining net should be the host for the sink terminal
  if(lat != sourcePNet->latency){
    //strcpy(tempStr, net->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
    tempStr = net->name + "PipelineNet" + itoa(net->pipelineNetList.size());
    currPNet = new PipeliningNet(tempStr, lat, net);
    net->pipelineNetList.push_back(currPNet);
  }
  else{
    currPNet = sourcePNet;
  }

  //Add the sink terminal to the current pipelining net
  currPNet->sinkVector.push_back(sinkTerminal);
  currPNet->linkDelayVector.push_back(current->rNode->pNetDelay.at(0).at(current->latency).at(current->exploreIndex));
  currPNet->slackVector.push_back(0.0);
  sinkTerminal->sourcePipeliningNet = currPNet;
  
  while(1){
    //See if we are done
    if(!current->rNode->pPredecessor.at(0).at(current->latency).at(current->exploreIndex).rNode){
      //Reattach this node to the source net
      //Find this node in the source net, then add the correct children/parent relationship
      //If a net has the same sink listed twice there will not be a previous tree node...
      if(prev_tree_node){
	curr_tree_node = currPNet->find_treeNode(current->rNode);
	curr_tree_node->addto_childList(prev_tree_node);
	prev_tree_node->set_parent(curr_tree_node);
      }

      //Go through the list of terminals to set the proper upstream costs
      currTerm = currPNet->source;

      //Release the beginning pExplore
      delete end;
      break;
    }
    
    if(current->rNode->type == DELAY){
      //This is the root of the current tree.
      //Notice that we do NOT use the pNetDelay value, this is the delay to the node from the source side, not the
      //delay starting from the node to the sink side.
      curr_tree_node = new treeNode(current->rNode, current->rNode->intrinsicDelay);
      
      //cout << "\t\tMaking new treeNode for ";
      //cout << "Node " << current->rNode->name << " delay " << current->rNode->intrinsicDelay << endl;
    }
    else{
      curr_tree_node = new treeNode(current->rNode, current->rNode->pNetDelay.at(0).at(current->latency).at(current->exploreIndex));

      //cout << "\t\tMaking new treeNode for ";
      //cout << "Node " << current->rNode->name << " delay " << current->rNode->pNetDelay.at(0).at(current->latency).at(current->exploreIndex) << endl;
    }
    
    if(prev_tree_node != 0){
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    //Increment occupancy
    current->rNode->set_occupancy(current->rNode->get_occupancy()+1);
    if(current->rNode->get_occupancy()>current->rNode->get_capacity()){
      sharingCost+=1;
      //cout << "Increasing sharing cost due to node " << current->rNode->name << endl;
    }
    if(current->rNode->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
    //Mark myself as in the PRouting tree
    if(current->rNode->check_inPRTree(current->latency)){
      ERROR("Node already in PRouting tree!");
    }
    else{
      current->rNode->set_inPRTree(current->latency);
    }

    //If we have hit a delay node so we need to do a few things
    if(current->rNode->type == DELAY){
      //1) Make it the root of the current pipelining net
      currPNet->rTree.set_root(curr_tree_node);

      //2) Make a new tree node for this delay node
      curr_tree_node = new treeNode(current->rNode, current->rNode->pNetDelay.at(0).at(current->latency).at(current->exploreIndex));

      //3) Evaluate the upstream pipelining net
      if(current->latency - 1 != sourcePNet->latency){
	//We need to create a new pipelining net and add it to the list
	//strcpy(tempStr, net->name);
	//sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
	tempStr = net->name + "PipelineNet" + itoa(net->pipelineNetList.size());
	nextPNet = new PipeliningNet(tempStr, current->latency - 1, net);
	net->pipelineNetList.push_back(nextPNet);
      }
      else{
	nextPNet = sourcePNet;
      }

      //4) Make a new terminal node for this delay node and add it to the main net's list
      //strcpy(tempStr, net->name);
      //sprintf(tempStr, "%sReg%d", tempStr, net->pipelineRegisterList.size());
      tempStr = net->name + "Reg" + itoa( net->pipelineRegisterList.size());
      currSink = new Terminal(net, current->rNode, nextPNet, currPNet, tempStr);
      net->pipelineRegisterList.push_back(currSink);

      //5) Make this terminal to be the source of the old net and add it to the sink list of the new one
      currPNet->source = currSink;
      nextPNet->sinkVector.push_back(currSink);
      nextPNet->linkDelayVector.push_back(curr_tree_node->linkDelay);
      nextPNet->slackVector.push_back(0.0);

      //6) Make the new net the current net since we a completely done with the previous one
      currPNet = nextPNet;
    }

    prev_tree_node=curr_tree_node;      
    current = &(current->rNode->pPredecessor.at(0).at(current->latency).at(current->exploreIndex));
  }
}

//This prints out the path back to the beginning of the route
void btpr(pExplore *pNode){
  cout << "\t\tNode " << pNode->rNode->name << " delay " << pNode->rNode->pNetDelay.at(0).at(pNode->latency).at(pNode->exploreIndex);
  cout << " AC index 0";
  cout << " latency " << pNode->latency << " E Index " << pNode->exploreIndex << endl;

  if(pNode->rNode->pPredecessor.at(0).at(pNode->latency).at(pNode->exploreIndex).rNode){
    btpr(&(pNode->rNode->pPredecessor.at(0).at(pNode->latency).at(pNode->exploreIndex)));
  }
}

void checkEmptyRRGraph(bool checkInPRTree){
  //Go through the entire routing graph and make sure that all nodes are unexplored at all latencies
  for(list<routingNode*>::iterator rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->inrTree == true){
      ERROR("Found routing node inside routing tree before routing has begun!");
    }
    else if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }

    for(int i = 0; i < NUMASSUMEDCRIT; i++){
      for(int j = 0; j <= MAXLATENCY; j++){
	if((*rNodeIter)->pIsExplored.at(i).at(j) != 0){
	  ERROR("Found explored routing node before routing has begun!");
	}
      }
    }
    if((*rNodeIter)->nodeID != 0){
      cout << "Routing Node " << (*rNodeIter)->name << endl;
      ERROR("Routing node has non-zero ID!");
    }
    if(checkInPRTree){
      //Make sure that the inPRTree variables are all zero
      for(vector <unsigned int>::iterator vectIter = (*rNodeIter)->inPRTree.begin(); vectIter != (*rNodeIter)->inPRTree.end(); vectIter++){
	if((*vectIter) != 0)
	  ERROR("Found routing node inside pipelined routing tree before routing begun!");
      }
    }
  }
  
  //Make sure that MAXID is 0
  if(MAXID != 0){
    ERROR("ID counter is non-zero!");
  }
}

