// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of a routing node. There is a separate routing
node for each logic block and IO pad terminal. Further, every 
wire or bus segment in the FPGA architecture is also represented 
as a routing node.

We assume the Elmore delay model here. Thus, each wire segment
routing node has to have a resistance and a capacitance.  
****************************************************************/    
#ifndef ROUTING_NODE_H
#define ROUTING_NODE_H

#include "constants.h"
#include "error.h"
#include "pExplore.h"
#include "util.h"
#include "routingEdge.h"

class Terminal;
class Net;

class routingNode{
 public:
  string name;  //name of routing node
  string blockName; //name of block that routing node belongs to

  //Adding in information here so that routing nodes know who they belong to
  int bank;
  void setBank(int b){bank = b;};
  int getBank(){return bank;};

  enum e_routingNodeType type;  //type of routing node i.e. source, sink, wire/bus, logic block
             //input, or logic block output, delay
  double R,Cin,Cout,C;  //wire routing nodes have a resistance and capacitances
  //list<Terminal*> terminalList;  //each source/sink routing node corresponds to one or more
                                 //logic block or IO pad terminals
  //list<Terminal*>::iterator* terminal_itr;  //pointer to terminal list iterator
  double baseCost, historyCost,presentSharing;  //Pathfinder cost function parameters
  double pfac;  //tuning parameter for presentSharing...VPR router
  int occupancy;  //num nets currently using this node

  int newOccupancy;

  int capacity;  //capacity of this node
  bool isExplored;  //whether this routing node has been explored by a wavefront
  bool inrTree;  //whether this node is in the current route tree of the net being routed 
  double pathCost;  //current cost of path to this node from source of net -- Pathfinder

  list<routingEdge*> faninList;  //list of fanin routing edges
  list<routingEdge*> fanoutList;  //list of fanout routing edges
  list<routingEdge*>::iterator* fanin_itr;  //pointer to fanin list iterator
  list<routingEdge*>::iterator* fanout_itr;  //pointer to fanout list iterator
  routingNode* predecessor;  //the fanin routing node that was used to explore this node
  bool isRouted;  //only valid for sink routing nodes. true if the sink has already been routed.
  //int isRouted;  //only valid for sink or delay routing nodes.  Indicates if there is a latency N route to this node.
  bool counted;  //whether this node's sharing has already been included in cost calculations
  int row_low; int col_low; int row_high; int col_high;
  int magicNum;
  
  //Timing-driven routing stuff
  double intrinsicDelay;
  double netDelay; //current delay of net from source to this node

  //Pipeline routing stuff
  vector <vector <vector <pExplore> > > pPredecessor;
  vector <vector <vector <double> > > pPathCost;
  vector <vector <vector <double> > > pNetDelay;
  vector <vector <int> > pIsExplored;
  vector <unsigned int> inPRTree;

  int nodeID;
  //End my stuff

  routingNode(int MAXLATENCY){
    int i, j, k;

    if(MAXLATENCY < 0){
      ERROR("Something is wrong!  MAXLATENCY < 0!");
    }

/*     //Initialize my variables here */
/*     pPredecessor.resize(NUMASSUMEDCRIT); */
/*     pPathCost.resize(NUMASSUMEDCRIT); */
/*     pNetDelay.resize(NUMASSUMEDCRIT); */
/*     pIsExplored.resize(NUMASSUMEDCRIT); */
/*     for(i = 0; i < NUMASSUMEDCRIT; i++){ */
/*       pPredecessor.at(i).resize(MAXLATENCY + 1); */
/*       pPathCost.at(i).resize(MAXLATENCY + 1); */
/*       pNetDelay.at(i).resize(MAXLATENCY + 1); */
/*       pIsExplored.at(i).resize(MAXLATENCY + 1); */
/*       for(j = 0; j <= MAXLATENCY; j++){ */
/* 	pPredecessor.at(i).at(j).resize(KFUNCTION(j)); */
/* 	pPathCost.at(i).at(j).resize(KFUNCTION(j)); */
/* 	pNetDelay.at(i).at(j).resize(KFUNCTION(j)); */

/* 	for(k = 0; k < KFUNCTION(j); k++){ */
/* 	  pPredecessor.at(i).at(j).at(k).clear(); */
/* 	  pPathCost.at(i).at(j).at(k) = DBL_MAX; */
/* 	  pNetDelay.at(i).at(j).at(k) = DBL_MAX; */
/* 	} */
/* 	pIsExplored.at(i).at(j) = 0; */
/*       } */
/*     } */

/*     inPRTree.resize((MAXLATENCY/NUMVECTORBITS) + 1); */
/*     for(int i = 0; i < inPRTree.size(); i++){ */
/*       inPRTree.at(i) = 0; */
/*     } */

    nodeID = 0;

    //blockType = GBT_NO_TYPE;
    type=RN_NO_TYPE;
    R=Cin=Cout=C=0.;
    presentSharing=0.;
    baseCost=1.0;
    historyCost=1.0;    
    occupancy=capacity=0;
    isExplored=false;
    pathCost=DBL_MAX;
    //pathCost=1E6;
    fanin_itr=fanout_itr=0;//terminal_itr=0;
    pfac=1.0;
    predecessor=0;
    inrTree=false;
    isRouted= -1;
    counted=false;
    row_low=col_low=row_high=col_high=-1;
    magicNum=19111;
    
    intrinsicDelay = 0.0;
    netDelay = DBL_MAX;

    bank = -1;
  }

  ~routingNode(){
    if(magicNum!=19111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;

    //Delete the fanout edges of this routing node
    for(list<routingEdge*>::iterator fanoutIter = fanoutList.begin(); fanoutIter != fanoutList.end(); fanoutIter++){
      delete (*fanoutIter);
    }
  }

  void set_name(char* n);  //give this routing node a name
  void set_blockName(char* n);
  void set_type(e_routingNodeType t);  //set type of routing node
  void set_R(double r);  //set resistance of wire routing node
  void set_Cin(double c);  //set input cap of wire routing node
  void set_Cout(double c);  //set output cap of wire routing node
  void set_predecessor(routingNode* pr);  //set this node's wavefront predecessor

  void set_baseCost(double b);  //set base cost of this routing node

  //update history cost of this routing node
  //This function will be called after every placement iteration
  void update_historyCost(int iter){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    
    //History cost disregards first iteration, then starts to accumulate
    if(iter > 1){
      /*     if(during_placement){ */
      /*       if(occupancy > capacity) */
      /* 	historyCost = doubmax(1.0, 0.9*historyCost + 0.5); */
      /*       else */
      /* 	historyCost = doubmax(1.0, 0.9*historyCost); */
      /*     } */
      /*     else{ */
      historyCost += (double) max(0, (occupancy - capacity));
      /*   } */
    }
  }

  void update_pfac();  //update the tuning parameter for presentSharing...VPR router
  void set_occupancy(int occ);  //set the occupancy of this routing node
  void set_capacity(int c);  //set the capacity of this routing node
  void set_isExplored(bool e);  //whether this node has been explored by a wavefront
  void set_inrTree(bool r);//whether this node is in the current route tree of
                             //the net being routed
  //return true if this routing node is in the route tree of a net
  bool get_inrTree(){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    return inrTree;
  }

  void set_pathCost(double p);  //set the current path cost of this routing node
  void addto_faninList(routingEdge* e);  //add a routing edge to the fanin list   
  void addto_fanoutList(routingEdge* e);  //add a routing edge to the fanout list   
  void create_faninListItr();  //create a new iterator for faninList
  void remove_faninListItr();  //remove an already existing faninList iterator 

  //create a new iterator for fanoutList
  void create_fanoutListItr(){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    if(fanout_itr)
      ERROR("fanout iterator already exists...");
    fanout_itr=new list<routingEdge*>::iterator;
    *fanout_itr = fanoutList.begin();
  }

  //remove an already existing fanoutList iterator 
  void remove_fanoutListItr(){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    if(fanout_itr==0)
      ERROR("trying to delete NULL fanout iterator...");
    delete fanout_itr;
    fanout_itr=0;
  } 

  //get the current cost of this routing node
  double returnCongestionCost(int iter){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    if(iter==1){
      return get_baseCost();
    }
    else{
      return get_baseCost()*get_historyCost(iter)*get_presentSharing(iter);
    }
  }

  //get the current timing cost of this routing node
  double returnTimingCost(int iter, double linkCriticality, double edgeDelay){
#if ROUTER_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
    //We can actually get negative timing cost if the link is super-critical (criticality larger than 1.0)
    if((linkCriticality * (intrinsicDelay + edgeDelay)) + ((1.0 - linkCriticality) * returnCongestionCost(iter)) < 0){
      ERROR("Returning negative timing cost!");
    }
#endif
    return ((linkCriticality * (intrinsicDelay + edgeDelay)) + ((1.0 - linkCriticality) * returnCongestionCost(iter)));
  };

  string get_name();  //get the name of this routing node
  enum e_routingNodeType get_type();  //get type of routing node
  double get_R();  //get the resistance of wire routing node
  double get_Cin();  //get the input cap of wire routing node
  double get_Cout();  //get the output cap of wire routing node
  routingNode* get_predecessor();  // get this node's wavefront predecessor

  routingEdge* getnext_faninList();  //get next element from faninList
  routingEdge* getnext_fanoutList(){
    // Get next element from fanoutList
    //if(magicNum!=19111)
    //  ERROR("trying to access a deleted node...");
    
    if(*fanout_itr == fanoutList.end())
      return 0;
    else{
      routingEdge* ret_ptr = **fanout_itr;
      (*fanout_itr)++;
      return ret_ptr;
    }
  }

  //get base cost of this routing node
  double get_baseCost(){
#ifdef ROUTE_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    return baseCost;
  }

  //get current history cost of this routing node
  double get_historyCost(int iter){
#ifdef ROUTE_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    return historyCost;
  }

  // Get current present sharing cost of routing node
  double get_presentSharing(int iter){
    double pfac;
#ifdef ROUTE_DEBUG
    if(magicNum!=19111)
      ERROR("trying to access a deleted node...");
#endif
    /***** THIS WAS FPGA'05 presentSharing UPDATE MECHANISM *****/
    pfac = 0.5 * pow(2, iter - 1);
    presentSharing=1.0 + doubmax(0.,((double)(1.0+occupancy-capacity)) * (pfac));
    
    /*   if(timing_driven){ */
    /*     pfac = iter; */
    /*   } */
    /*   else{ */
    /*     pfac = pow(1.5,(double)iter); */
    /*     //pfac = iter; */
    /*   } */
    /*   if(pathfinder_pshare) */
    /*     presentSharing=1.0 + doubmax(0.,((double)(1.0+occupancy-capacity)) * (pfac)); */
    /*   else */
    /*     presentSharing=1.0 + doubmax(0.,(double)(1.0+occupancy-capacity)); */
    
    return presentSharing;
  }

  int get_occupancy();  //get the current occupancy of this node
  int get_capacity();  //get the capacity of this node
  bool get_isExplored();  //get whether this node has been explored by a wavefront
  double get_pathCost();  //get current path cost of this routing node
  //bool foundTerminal(Net* net);  //applicable only to sink routing nodes. if a sink 
                                //terminal of net is mapped to this routing node return true 
  void print_info();  //print details about this routing node
  //void clear_terminalList();  //clear the list of terminals mapped to this routing node
  void set_counted(bool cn);  //whether this node's sharing has already been included in 
                                //cost calculations 
  bool get_counted();  //true if this node's sharing has been considered in cost calculations

  void set_inPRTree(int latency);
  bool check_inPRTree(int latency);
  bool check_exploredAtLatency(int latency);
};


#endif


