// Copyright (C) 2008 Ken Eguro & Scott Hauck
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
This file contains class definitions of the VPR FPGA architecture
and the individual clbs that constitute a VPR-style FPGA. Each
vprCLB object holds information about the source/sink/input-pin/
output-pin routing nodes that it provides. Each vprCLB
owns the track segments in the channels that are north and east
of it. Further, the vprLUT also holds a pointer to the logic
block that is currently placed on it.
****************************************************************/    
#ifndef VPR_ARCH_H
#define VPR_ARCH_H

class logicBlock;
class Netlist;
class c_clb;
class vprCLB;

//This represents a bank of registers that have exactly 1 input and one output
class registerBank{
 public:
  vector <routingNode *> routingNodeVector;
  list <Terminal *> currSinkList;
  list <Terminal *> currSourceList;
  vprCLB* hostCLB;
  int hostBankNum;
  logicBlock *currBlock;


  registerBank(){
    hostBankNum = -1;
    hostCLB = 0;
    currBlock = 0;
  };
  ~registerBank(){};
  void setHostCLB(vprCLB *clb, int bankNum){hostCLB = clb; hostBankNum = bankNum;};
  vprCLB* getHostCLB(){return hostCLB;};
  int getHostBankNum(){return hostBankNum;};
  
  void addRoutingNode(routingNode *r);
  Terminal* getFirstSink();
  Terminal* getLastSource();
  routingNode* getFirstRoutingNode();
  //Is this register bank empty?
  bool isBankEmpty();
  int getNumMappedRegs();
  int getNumPossibleDelay();
  int getNumFreeRegs();
  bool bankFull();

  void placeTerminal(Terminal *T);
  void placeTerminalFront(Terminal *T);
  void placeLogicBlock(logicBlock *lb);
  void unPlaceLogicBlock(logicBlock *lb);
  void insertRegisterForwards(Terminal *T, Terminal *parent);
  void insertRegisterBackwards(Terminal *T, Terminal *parent);
  void swapBank(registerBank *otherBank);
  void copyContentsToAndEmpty(registerBank *otherBank);
  void updateSinkDelaysAndSlacks();
  void updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay);
  double determineDeltaWireCost();
  double determineDeltaTimingCost();
  double determineDeltaTimingCostPrime();

  void migrateFirstNRegs(registerBank *otherBank, int numRegs);
  void migrateLastNRegs(registerBank *otherBank, int numRegs);
  
  void duplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  void zipRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  void reDuplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  
  void removeStartReference(Terminal *T);
  void removeEndReference(Terminal *T);

  void freeBank();

  double returnTotalWireCost();
  double returnTotalTimingCost();
  void updateDeletedCutCost();
  
  void print_info();

  void updateRoutingNodes();

  logicBlock* get_currBlock(){
    return currBlock;
  }

  //Are the LUTs and other registers banks in the clb at the same location empty?
  bool emptyCLB();
  
  void swapMapping(logicBlock* myLB, registerBank* otherBank, logicBlock* otherLB);
  //void swapEntireCLBMappings(logicBlock* myLB, registerBank* otherBank, logicBlock* otherLB);
  void swapEntireCLBMappings(registerBank* otherBank);

  void updateCostAndDeltasForEntireCLB(registerBank* otherRegisterBank, 
				       bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForEntireCLBCostDeltaDelay(registerBank* otherRegisterBank, bool timing_driven, 
						     double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime);

  c_clb* returnMappedCLB();

  registerBank *getRegBankInParentHostCLB();
};


class vprBaseBlock{
 public:
  int row,col;  //row and col coordinates of this block
  int magicNum;

  vprBaseBlock(int r,int c);
  ~vprBaseBlock();
  int get_row();  //get the row coord of this block
  int get_col();  //get the col coord of this block
  virtual void print_info();  //print details about this block
  //void print_info();  //print details about this block
};


/****************************************************************
This class represents a LUT in the VPR architecture.
****************************************************************/
class vprLUT : public vprBaseBlock{
 public:
  routingNode* sourceNode;  //source routing node
  routingNode* sinkNode;  //sink routing node
  routingNode* clock_sinkNode;  //sink routing node for clock routing...NOT USED!!
  logicBlock* currBlock;  //the logic block that is currently placed at this LUT
  enum e_genericBlockType blockType;
  int derived_magicNum;

  vprLUT(int r,int c);
  ~vprLUT();

  void set_currBlock(logicBlock* l);  //set the logic block that is placed at this CLB
  logicBlock* get_currBlock();  //get the logic block that is placed at this CLB
  void set_sourceNode(routingNode* r);  //set the source node of this CLB
  void set_sinkNode(routingNode* r);  //set the sink node of this CLB
  routingNode* get_sourceNode();  //get the source node of this CLB
  routingNode* get_sinkNode();  //get the sink node of this CLB
  void set_blockType(enum e_genericBlockType t);
  enum e_genericBlockType get_blockType();
  void print_info();  //print details about this logic block
  void clearCurrBlock();

  //Is the CLB that this LUT belongs to empty? (LUTs and registers)
  bool emptyCLB();

  //Are the LUTs in this CLB empty?
  bool emptyCLBLUTS();

  //Is this LUT empty?
  bool isEmpty(){
    if(currBlock)
      return false;
    else
      return true;
  };

  void swapMapping(int myIndex, logicBlock* myLB, vprLUT* otherLUT, int otherIndex, logicBlock* otherLB);
  void swapStickyMapping(vprLUT* otherLUT);
  void swapEntireCLBMappings(vprLUT* otherLUT);

  void updateCostAndDeltasForAllCLBLUTs(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForAllCLBLUTsCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, 
						      double *deltaTiming, double *deltaTimingPrime);

  void updateCostAndDeltasForEntireCLB(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForEntireCLBCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, 
						     double *deltaTiming, double *deltaTimingPrime);

  c_clb* returnMappedCLBFromLUTsAndFF();
  c_clb* returnMappedCLBFromLUTs();
};

//This represents a group of 1 or more LUTs & register banks
class vprCLB :public vprBaseBlock{
 public:
  vector<registerBank *> registerBankVector;
  vector<vprLUT *> lutVector;
  int derived_magicNum;

  vprCLB(int row, int col);
  ~vprCLB(){
    //Delete the luts and registers
    for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
      delete *rbIter;
    }
    for(vector<vprLUT *>::iterator lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
      delete *lutIter;
    }
  }

  void addAllRegBanksToRegisterBankVector();
  void addAllRegBanksToFreeRegisterBankList();

  void addRoutingNodeToRegisterBank(routingNode* r, int bank);
  
  int getNumLUTs(){return lutVector.size();};
  int getNumRegBanks(){return registerBankVector.size();};
  
  registerBank* getRegBank(int i){
    return registerBankVector.at(i);
  }

  vprLUT* getLUT(int i){
    return lutVector.at(i);
  }

  vprLUT* getOrMakeLUT(int i);

  //Are all of the register banks in this delay block empty?
  bool areRegistersEmpty();
  bool areLUTsEmpty();

  registerBank* getFirstEmptyUnrelatedRegisterBank();
  registerBank* getFirstEmptyRegisterBank();

  void swapAllFFs(vprCLB *otherCLB);
  void updateSinkDelaysAndSlacksForAllFFs();
  void updateCostAndDeltasForAllFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForAllFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime);
};

/****************************************************************
This class represents an IOB in the VPR architecture. Note that
there are multiple IO elements in a single IOB, and hence there
are multiple source and sink nodes in each IOB.
****************************************************************/
class vprIOB : public vprBaseBlock{
 public:
  vector<routingNode*>* sourceArray;  //ptr to array of source routing nodes
  vector<routingNode*>* sinkArray;  //ptr to array of sink routing nodes
  vector<logicBlock*>* currBlockArray;  //ptr to array of logic blocks placed at this IOB
  int numIOs;  //number of IOs per IOB
  int derived_magicNum;

  vprIOB(int r,int c);
  vprIOB(int r,int c,int n);
  ~vprIOB();
  void addto_sourceArray(routingNode* r,int i);  //add source node to sourceArray at index 'i'
  void addto_sinkArray(routingNode* r,int i);  //add sink node to sinkArray at index 'i'
  void addto_currBlockArray(logicBlock* l,int i);  //place logic block at 'ith' IO location 
  routingNode* getfrom_sourceArray(int i);  //get node in sourceArray at index 'i'
  routingNode* getfrom_sinkArray(int i);  //get node in sinkArray at index 'i'
  logicBlock* getfrom_currBlockArray(int i);  //get logic block placed at 'ith' IO location
  void print_info();  //print details about this IOB
  int get_numIOs(); //get the number of IOs in this IOB
  void lineup_IOs();
  void clearCurrBlockArray(int i);
  int getRandomIndex(){
    return (lrand48() % numIOs);
  };

  void swapMapping(int myIndex, logicBlock* myLB, vprIOB* otherIOB, int otherIndex, logicBlock* otherLB);
};


class vprFPGA{
 public:
  int numRows,numCols;  //num CLB rows and columns respectively
  int numCLBs,numIOs;  //num CLBs and IOBs in this architecture
  int numIOs_perIOB;  //number of separate IOs per IOB
  //vector<vector<void*>*>* blockArray;  //2-D array of CLBs and IOBs
  vector<vector<void*> > array;  //2-D array of CLBs and IOBs
  
  list<routingNode*> rnodeList;  //List that holds ALL routing nodes in the architecture.
                                 //This list will be populated from the rr_graph.echo file
                                 //dumped by VPR. The routing edges that connect nodes are
                                 //'owned' by the routing nodes at which the edges originate.
  list<routingNode*>::iterator* rnode_itr;  //iterator for list of routing nodes

  int magicNum;

  vprFPGA(int num_rows,int num_cols,int num_ios);
  ~vprFPGA();

  //void* get_blockAt(int r,int c);  //get block at location c,r
  int get_numRows(){
#ifdef PLACER_DEBUG
    if(magicNum!=23111)
      ERROR("trying to double delete node here...");
#endif
    return numRows;
  }

  int get_numCols(){
#ifdef PLACER_DEBUG
    if(magicNum!=23111)
      ERROR("trying to double delete node here...");
#endif
    return numCols;
  }

  int get_numCLBs();  //get num CLBs in this architecture
  int get_numIOs();  //get num IOBs in this architecture
  int countLUT();
  int countMult();
  int countRegBanks();
  //int countRegTotal();
  int countRAM();
  void addto_rnodeList(routingNode* r);  //add routing nodes to rnodeList
  routingNode* getnext_rnodeList();  //get next routing node in rnodeList
  void create_rnodeListItr();  //create an iterator for rnodeList
  void remove_rnodeListItr();  //remove iterator for rnodeList
  //routingNode* found_routingNode(char* rname);  //look for routing node with rname and return

  //count (and optionally print) the total sharing of routing nodes in this FPGA
  //long int countSharing(int opt);

  //count (and optionally print) the total occupancy of congested routing nodes in this FPGA
  long int countCongestedNodeOccupancy(int opt);

  //update the history costs of all routing nodes in this FPGA
  void updateHistoryCosts(int iter);

  void print_info();  //print arch info about this FPGA

  void lineup_IOBs();  //line up IOs in each IOB's vectors 

  //Timing-driven routing stuff
  double calculateAverageIntrinsicDelay();
  void scaleBaseCosts(double averageIntrinsicDelay);

  //Pipeline stuff
  void initPlacementDS();

  int countDelayReg();
  void zeroOccupancy();
  long int countNewSharing(int opt);

  //Pipeline placement
  void create_block_type_lists();
  //void setupDelayBlocks();

  vector <vector <vector <vector <vector <vector <double> > > > > > *timingTable;
  vector <vector <vector <vector <vector <vector <int> > > > > > *timingNumHitTable;
  //double *timingTable;
  //int *timingNumHitTable;

  void createTimingTable();
  void initializeTimingTable();
  void initializeTimingNumHitTable();
  void calculateTimingTable();
  void emptyTimingTable();
  void emptyTimingNumHitTable();

  vector <double> cutCostVector;  
  void createCutCostVector();
  void clearCutCostVector();

  void sweepTimingIOPins();
  void sweepTimingLogic();
  void sweepTimingDelay();
  void findAllSinksFrom(void* sourceBlock, e_timingMoveBlockType sourceType, int sourceBank, int sourceRow, int sourceCol);
  void addTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  void incNumHitTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  void setTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  double getTimingTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
    //return timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1);
    
    //These changes assume that all banks have the same connectivity    
    //if(timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0){
    //  cout << "Something went wrong!" << endl;
    //}
    return timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1);
  }

  vector<vprIOB *> iobList;
  vector<vprLUT *> lutList;
  //I'll convert this to a vprCLB later
  vector<vprLUT *> multList;
  vector<vprLUT *> ramList;

  //vector<vector <delayBlock *> > delayBlockArray;
  vector<registerBank *> registerBankVector;
  list<registerBank *> freeRegisterBankList;
  int numFreeRegisters;

  //delayBlock* getOrMakeDelayBlockAt(int row, int col);
  //delayBlock* getDelayBlockAt(int row, int col);
  vprIOB* getOrMakeIOBAt(int row, int col);
  vprIOB* getIOBAt(int row, int col);
  vprCLB* getOrMakeCLBAt(int row, int col);
  vprCLB* getCLBAt(int row, int col);

  int getNumberofBlockType(int blockType);
  
  vprIOB* getRandomIOB(){return iobList.at(lrand48() % iobList.size());};
  vprLUT* getRandomLUT(){return lutList.at(lrand48() % lutList.size());};
  vprLUT* getRandomMult(){return multList.at(lrand48() % multList.size());};
  vprLUT* getRandomRam(){return ramList.at(lrand48() % ramList.size());};

  registerBank * getRandomRegBank(){
    return registerBankVector.at(lrand48() % registerBankVector.size());
  };
  registerBank * getRandomFreeRegBank();

  vprIOB* getOtherIOB(int row, int col, int rLim);
  vprLUT* getOtherLUT(int row, int col, int rLim);
  //vprLUT* getOtherMult(int row, int col, int rLim);
  //vprLUT* getOtherRam(int row, int col, int rLim);
  registerBank* getOtherRegBank(int row, int col, int rLim);

  void occupyRegBank(registerBank* rb);
  void vacateRegBank(registerBank* rb);

  void checkFreeRegisters();

  void placeClockPin(int *clockRow, int *clockCol, int *clockPTC);

  void checkArrayMappings();
};

#endif

